@@ -1,6 +1,6 @@
 /*
  *   stunnel       Universal SSL tunnel
- *   Copyright (C) 1998-2009 Michal Trojnara <Michal.Trojnara@mirt.net>
+ *   Copyright (C) 1998-2011 Michal Trojnara <Michal.Trojnara@mirt.net>
  *
  *   This program is free software; you can redistribute it and/or modify it
  *   under the terms of the GNU General Public License as published by the
@@ -38,29 +38,25 @@
 #include "common.h"
 #include "prototypes.h"
 
-#ifndef NO_RSA
-
-/* Cache temporary keys up to 2048 bits */
-#define KEY_CACHE_LENGTH 2049
-
-/* Cache temporary keys up to 1 hour */
-#define KEY_CACHE_TIME 3600
-
-#endif /* NO_RSA */
-
 /**************************************** prototypes */
 
-/* RSA/DH initialization */
-#ifndef NO_RSA
-static RSA *tmp_rsa_cb(SSL *, int, int);
-static RSA *make_temp_key(int);
-#endif /* NO_RSA */
-#ifdef USE_DH
-static int init_dh(SSL_CTX *, LOCAL_OPTIONS *);
-#endif /* USE_DH */
+/* SNI */
+#ifndef OPENSSL_NO_TLSEXT
+static int servername_cb(SSL *, int *, void *);
+#endif
+
+/* DH/ECDH initialization */
+#ifndef OPENSSL_NO_DH
+static int init_dh(SERVICE_OPTIONS *);
+static DH *read_dh(char *);
+static DH *get_dh2048(void);
+#endif /* OPENSSL_NO_DH */
+#ifndef OPENSSL_NO_ECDH
+static int init_ecdh(SERVICE_OPTIONS *);
+#endif /* USE_ECDH */
 
 /* loading certificate */
-static void load_certificate(LOCAL_OPTIONS *);
+static int load_certificate(SERVICE_OPTIONS *);
 static int password_cb(char *, int, int, void *);
 
 /* session cache callbacks */
@@ -73,70 +69,46 @@ static void cache_transfer(SSL_CTX *, const unsigned int, const unsigned,
     unsigned char **, unsigned int *);
 
 /* info callbacks */
-#if SSLEAY_VERSION_NUMBER >= 0x00907000L
 static void info_callback(const SSL *, int, int);
-#else /* OpenSSL-0.9.7 */
-static void info_callback(SSL *, int, int);
-#endif /* OpenSSL-0.9.7 */
 static void print_stats(SSL_CTX *);
 
-static void sslerror_stack(void);
+static void sslerror_queue(void);
+static void sslerror_log(unsigned long, char *);
 
 /**************************************** initialize section->ctx */
 
-void context_init(LOCAL_OPTIONS *section) { /* init SSL context */
-    struct stat st; /* buffer for stat */
-
-    /* check if certificate exists */
-    if(!section->key) /* key file not specified */
-        section->key=section->cert;
-#ifdef HAVE_OSSL_ENGINE_H
-    if(!section->engine)
-#endif
-    if(section->option.cert) {
-        if(stat(section->key, &st)) {
-            ioerror(section->key);
-            die(1);
-        }
-#if !defined(USE_WIN32) && !defined(USE_OS2)
-        if(st.st_mode & 7)
-            s_log(LOG_WARNING, "Wrong permissions on %s", section->key);
-#endif /* defined USE_WIN32 */
-    }
+int context_init(SERVICE_OPTIONS *section) { /* init SSL context */
+    s_log(LOG_INFO, "Initializing SSL context for service %s",
+        section->servname);
 
     /* create SSL context */
     if(section->option.client)
         section->ctx=SSL_CTX_new(section->client_method);
-    else /* Server mode */
+    else /* server mode */
         section->ctx=SSL_CTX_new(section->server_method);
     SSL_CTX_set_ex_data(section->ctx, opt_index, section); /* for callbacks */
-    if(!section->option.client) { /* RSA/DH callbacks */
-#ifndef NO_RSA
-        SSL_CTX_set_tmp_rsa_callback(section->ctx, tmp_rsa_cb);
-#endif /* NO_RSA */
-#ifdef USE_DH
-        if(init_dh(section->ctx, section))
-            s_log(LOG_WARNING, "Diffie-Hellman initialization failed");
-#endif /* USE_DH */
-    }
-    if(section->ssl_options) {
-        s_log(LOG_DEBUG, "Configuration SSL options: 0x%08lX",
-            section->ssl_options);
-        s_log(LOG_DEBUG, "SSL options set: 0x%08lX",
-            SSL_CTX_set_options(section->ctx, section->ssl_options));
-    }
-    if(section->cipher_list) {
-        if (!SSL_CTX_set_cipher_list(section->ctx, section->cipher_list)) {
-            sslerror("SSL_CTX_set_cipher_list");
-            die(1);
-        }
-    }
-#if SSLEAY_VERSION_NUMBER >= 0x00906000L
-    SSL_CTX_set_mode(section->ctx,
-        SSL_MODE_ENABLE_PARTIAL_WRITE|SSL_MODE_ACCEPT_MOVING_WRITE_BUFFER);
-#endif /* OpenSSL-0.9.6 */
 
-    /* session cache */
+    /* initialize certificate verification */
+    if(load_certificate(section))
+        return 1; /* FAILED */
+    if(verify_init(section))
+        return 1; /* FAILED */
+
+    /* initialize DH/ECDH server mode */
+    if(!section->option.client) {
+#ifndef OPENSSL_NO_TLSEXT
+        SSL_CTX_set_tlsext_servername_arg(section->ctx, section);
+        SSL_CTX_set_tlsext_servername_callback(section->ctx, servername_cb);
+#endif /* OPENSSL_NO_TLSEXT */
+#ifndef OPENSSL_NO_DH
+        init_dh(section); /* ignore the result (errors are not critical) */
+#endif /* OPENSSL_NO_DH */
+#ifndef OPENSSL_NO_ECDH
+        init_ecdh(section); /* ignore the result (errors are not critical) */
+#endif /* OPENSSL_NO_ECDH */
+    }
+
+    /* setup session cache */
     SSL_CTX_set_session_cache_mode(section->ctx, SSL_SESS_CACHE_BOTH);
     SSL_CTX_set_timeout(section->ctx, section->session_timeout);
     if(section->option.sessiond) {
@@ -145,139 +117,203 @@ void context_init(LOCAL_OPTIONS *section) { /* init SSL context */
         SSL_CTX_sess_set_remove_cb(section->ctx, sess_remove_cb);
     }
 
-    /* info callback */
+    /* set info callback */
     SSL_CTX_set_info_callback(section->ctx, info_callback);
 
-    /* initialize certificate verification */
-    if(section->option.cert)
-        load_certificate(section);
-    verify_init(section);
-
-    s_log(LOG_DEBUG, "SSL context initialized for service %s",
-        section->servname);
+    /* ciphers, options, mode */
+    if(section->cipher_list)
+        if(!SSL_CTX_set_cipher_list(section->ctx, section->cipher_list)) {
+            sslerror("SSL_CTX_set_cipher_list");
+            return 1; /* FAILED */
+        }
+    s_log(LOG_DEBUG, "SSL options set: 0x%08lX",
+        SSL_CTX_set_options(section->ctx, section->ssl_options));
+    SSL_CTX_set_mode(section->ctx,
+        SSL_MODE_ENABLE_PARTIAL_WRITE|SSL_MODE_ACCEPT_MOVING_WRITE_BUFFER);
+    s_log(LOG_INFO, "SSL context initialized");
+    return 0; /* OK */
 }
 
-/**************************************** temporary RSA keys generation */
+/**************************************** SNI callback */
 
-#ifndef NO_RSA
+#ifndef OPENSSL_NO_TLSEXT
 
-static RSA *tmp_rsa_cb(SSL *s, int export, int keylen) {
-    static int initialized=0;
-    static struct keytabstruct {
-        RSA *key;
-        time_t timeout;
-    } keytable[KEY_CACHE_LENGTH];
-    static RSA *longkey=NULL;
-    static int longlen=0;
-    static time_t longtime=0;
-    RSA *oldkey, *retval;
-    time_t now;
-    int i;
+static int servername_cb(SSL *ssl, int *ad, void *arg) {
+    SERVICE_OPTIONS *section=(SERVICE_OPTIONS *)arg;
+    const char *servername=SSL_get_servername(ssl, TLSEXT_NAMETYPE_host_name);
+    SERVERNAME_LIST *list;
+    CLI *c;
 
-    enter_critical_section(CRIT_KEYGEN);
-        /* only one make_temp_key() at a time */
-    if(!initialized) {
-        for(i=0; i<KEY_CACHE_LENGTH; i++) {
-            keytable[i].key=NULL;
-            keytable[i].timeout=0;
+    /* leave the alert type at SSL_AD_UNRECOGNIZED_NAME */
+    (void)ad; /* skip warning about unused parameter */
+    if(!section->servername_list_head) /* no virtual services defined */
+        return SSL_TLSEXT_ERR_OK;
+    if(!servername) /* no SNI extension received from the client */
+        return SSL_TLSEXT_ERR_NOACK;
+
+    for(list=section->servername_list_head; list; list=list->next)
+        if(!strcasecmp(servername, list->servername)) {
+            c=SSL_get_ex_data(ssl, cli_index);
+            c->opt=list->opt;
+            SSL_set_SSL_CTX(ssl, c->opt->ctx);
+            s_log(LOG_NOTICE, "SNI: switched to section %s",
+                c->opt->servname);
+#ifdef USE_LIBWRAP
+            libwrap_auth(c); /* retry on a service switch */
+#endif /* USE_LIBWRAP */
+            return SSL_TLSEXT_ERR_OK;
         }
-        initialized=1;
-    }
-    time(&now);
-    if(keylen<KEY_CACHE_LENGTH) {
-        if(keytable[keylen].timeout<now) {
-            oldkey=keytable[keylen].key;
-            keytable[keylen].key=make_temp_key(keylen);
-            keytable[keylen].timeout=now+KEY_CACHE_TIME;
-            if(oldkey)
-                RSA_free(oldkey);
-        }
-        retval=keytable[keylen].key;
-    } else { /* temp key > 2048 bits.  Is it possible? */
-        if(longtime<now || longlen!=keylen) {
-            oldkey=longkey;
-            longkey=make_temp_key(keylen);
-            longtime=now+KEY_CACHE_TIME;
-            longlen=keylen;
-            if(oldkey)
-                RSA_free(oldkey);
-        }
-        retval=longkey;
-    }
-    leave_critical_section(CRIT_KEYGEN);
-    return retval;
+    s_log(LOG_ERR, "SNI: no service defined for server %s", servername);
+    return SSL_TLSEXT_ERR_ALERT_FATAL;
 }
+/* TLSEXT callback return codes:
+ *  - SSL_TLSEXT_ERR_OK
+ *  - SSL_TLSEXT_ERR_ALERT_WARNING
+ *  - SSL_TLSEXT_ERR_ALERT_FATAL
+ *  - SSL_TLSEXT_ERR_NOACK */
 
-static RSA *make_temp_key(int keylen) {
-    RSA *result;
-
-    s_log(LOG_DEBUG, "Generating %d bit temporary RSA key...", keylen);
-#if SSLEAY_VERSION_NUMBER >= 0x0900
-    result=RSA_generate_key(keylen, RSA_F4, NULL, NULL);
-#else
-    result=RSA_generate_key(keylen, RSA_F4, NULL);
-#endif
-    s_log(LOG_DEBUG, "Temporary RSA key created");
-    return result;
-}
-
-#endif /* NO_RSA */
+#endif /* OPENSSL_NO_TLSEXT */
 
 /**************************************** DH initialization */
 
-#ifdef USE_DH
-static int init_dh(SSL_CTX *ctx, LOCAL_OPTIONS *section) {
-    FILE *fp;
-    DH *dh;
-    BIO *bio;
+#ifndef OPENSSL_NO_DH
 
-    fp=fopen(section->cert, "r");
-    if(!fp) {
-#ifdef USE_WIN32
-        /* fopen() does not return the error via GetLastError() on Win32 */
-        s_log(LOG_ERR, "Failed to open %s", section->cert);
-#else
-        ioerror(section->cert);
-#endif
-        return -1; /* FAILED */
+static int init_dh(SERVICE_OPTIONS *section) {
+    DH *dh;
+
+    dh=read_dh(section->cert);
+    if(!dh)
+        dh=get_dh2048();
+    if(!dh) {
+        s_log(LOG_NOTICE, "DH initialization failed");
+        return 1; /* FAILED */
     }
-    bio=BIO_new_fp(fp, BIO_CLOSE|BIO_FP_TEXT);
-    if(!bio) {
-        s_log(LOG_ERR, "BIO_new_fp failed");
-        return -1; /* FAILED */
-    }
-    if((dh=PEM_read_bio_DHparams(bio, NULL, NULL
-#if SSLEAY_VERSION_NUMBER >= 0x00904000L
-            , NULL
-#endif
-            ))) {
-        BIO_free(bio);
-        s_log(LOG_DEBUG, "Using Diffie-Hellman parameters from %s",
-            section->cert);
-    } else { /* failed to load DH parameters from file */
-        BIO_free(bio);
-        s_log(LOG_NOTICE, "Could not load DH parameters from %s", section->cert);
-        return -1; /* FAILED */
-    }
-    SSL_CTX_set_tmp_dh(ctx, dh);
-    s_log(LOG_INFO, "Diffie-Hellman initialized with %d bit key",
-        8*DH_size(dh));
+    SSL_CTX_set_tmp_dh(section->ctx, dh);
+    s_log(LOG_DEBUG, "DH initialized with %d-bit key", 8*DH_size(dh));
     DH_free(dh);
     return 0; /* OK */
 }
-#endif /* USE_DH */
+
+static DH *read_dh(char *cert) {
+    DH *dh;
+    BIO *bio;
+
+    if(!cert) {
+        s_log(LOG_DEBUG, "No certificate available to load DH parameters");
+        return NULL; /* FAILED */
+    }
+    bio=BIO_new_file(cert, "r");
+    if(!bio) {
+        sslerror("BIO_new_file");
+        return NULL; /* FAILED */
+    }
+    dh=PEM_read_bio_DHparams(bio, NULL, NULL, NULL);
+    BIO_free(bio);
+    if(!dh) {
+        while(ERR_get_error())
+            ; /* OpenSSL error queue cleanup */
+        s_log(LOG_DEBUG, "Could not load DH parameters from %s", cert);
+        return NULL; /* FAILED */
+    }
+    s_log(LOG_DEBUG, "Using DH parameters from %s", cert);
+    return dh;
+}
+
+static DH *get_dh2048() {
+    static unsigned char dh2048_p[]={ /* OpenSSL DH parameters */
+        0xED,0x92,0x89,0x35,0x82,0x45,0x55,0xCB,0x3B,0xFB,0xA2,0x76,
+        0x5A,0x69,0x04,0x61,0xBF,0x21,0xF3,0xAB,0x53,0xD2,0xCD,0x21,
+        0xDA,0xFF,0x78,0x19,0x11,0x52,0xF1,0x0E,0xC1,0xE2,0x55,0xBD,
+        0x68,0x6F,0x68,0x00,0x53,0xB9,0x22,0x6A,0x2F,0xE4,0x9A,0x34,
+        0x1F,0x65,0xCC,0x59,0x32,0x8A,0xBD,0xB1,0xDB,0x49,0xED,0xDF,
+        0xA7,0x12,0x66,0xC3,0xFD,0x21,0x04,0x70,0x18,0xF0,0x7F,0xD6,
+        0xF7,0x58,0x51,0x19,0x72,0x82,0x7B,0x22,0xA9,0x34,0x18,0x1D,
+        0x2F,0xCB,0x21,0xCF,0x6D,0x92,0xAE,0x43,0xB6,0xA8,0x29,0xC7,
+        0x27,0xA3,0xCB,0x00,0xC5,0xF2,0xE5,0xFB,0x0A,0xA4,0x59,0x85,
+        0xA2,0xBD,0xAD,0x45,0xF0,0xB3,0xAD,0xF9,0xE0,0x81,0x35,0xEE,
+        0xD9,0x83,0xB3,0xCC,0xAE,0xEA,0xEB,0x66,0xE6,0xA9,0x57,0x66,
+        0xB9,0xF1,0x28,0xA5,0x3F,0x22,0x80,0xD7,0x0B,0xA6,0xF6,0x71,
+        0x93,0x9B,0x81,0x0E,0xF8,0x5A,0x90,0xE6,0xCC,0xCA,0x6F,0x66,
+        0x5F,0x7A,0xC0,0x10,0x1A,0x1E,0xF0,0xFC,0x2D,0xB6,0x08,0x0C,
+        0x62,0x28,0xB0,0xEC,0xDB,0x89,0x28,0xEE,0x0C,0xA8,0x3D,0x65,
+        0x94,0x69,0x16,0x69,0x53,0x3C,0x53,0x60,0x13,0xB0,0x2B,0xA7,
+        0xD4,0x82,0x87,0xAD,0x1C,0x72,0x9E,0x41,0x35,0xFC,0xC2,0x7C,
+        0xE9,0x51,0xDE,0x61,0x85,0xFC,0x19,0x9B,0x76,0x60,0x0F,0x33,
+        0xF8,0x6B,0xB3,0xCA,0x52,0x0E,0x29,0xC3,0x07,0xE8,0x90,0x16,
+        0xCC,0xCC,0x00,0x19,0xB6,0xAD,0xC3,0xA4,0x30,0x8B,0x33,0xA1,
+        0xAF,0xD8,0x8C,0x8D,0x9D,0x01,0xDB,0xA4,0xC4,0xDD,0x7F,0x0B,
+        0xBD,0x6F,0x38,0xC3,};
+    static unsigned char dh2048_g[]={0x02,};
+    DH *dh;
+
+    dh=DH_new();
+    if(!dh)
+        return NULL;
+    dh->p=BN_bin2bn(dh2048_p, sizeof dh2048_p, NULL);
+    dh->g=BN_bin2bn(dh2048_g, sizeof dh2048_g, NULL);
+    if(!dh->p || !dh->g) {
+        DH_free(dh);
+        return NULL;
+    }
+    s_log(LOG_DEBUG, "Using hardcoded DH parameters");
+    return dh;
+}
+
+#endif /* OPENSSL_NO_DH */
+
+/**************************************** ECDH initialization */
+
+#ifndef OPENSSL_NO_ECDH
+static int init_ecdh(SERVICE_OPTIONS *section) {
+    EC_KEY *ecdh;
+
+    ecdh=EC_KEY_new_by_curve_name(section->curve);
+    if(!ecdh) {
+        s_log(LOG_ERR, "Unable to create curve %s",
+            OBJ_nid2ln(section->curve));
+        return 1; /* FAILED */
+    }
+    SSL_CTX_set_tmp_ecdh(section->ctx, ecdh);
+    EC_KEY_free(ecdh);
+    s_log(LOG_DEBUG, "ECDH initialized with curve %s",
+        OBJ_nid2ln(section->curve));
+    return 0; /* OK */
+}
+#endif /* OPENSSL_NO_ECDH */
 
 /**************************************** loading certificate */
 
 static int cache_initialized=0;
 
-static void load_certificate(LOCAL_OPTIONS *section) {
+static int load_certificate(SERVICE_OPTIONS *section) {
     int i, reason;
     UI_DATA ui_data;
 #ifdef HAVE_OSSL_ENGINE_H
     EVP_PKEY *pkey;
-    UI_METHOD *uim;
+    UI_METHOD *ui_method;
 #endif
+    struct stat st; /* buffer for stat */
+
+    /* check if certificate exists */
+    if(!section->key) /* key file not specified */
+        section->key=section->cert;
+#ifdef HAVE_OSSL_ENGINE_H
+    if(!section->engine)
+#endif
+    if(section->key) {
+        if(stat(section->key, &st)) {
+            ioerror(section->key);
+            return 1; /* FAILED */
+        }
+#if !defined(USE_WIN32) && !defined(USE_OS2)
+        if(st.st_mode & 7)
+            s_log(LOG_WARNING, "Insecure file permissions on %s",
+                section->key);
+#endif /* defined USE_WIN32 */
+    }
+
+    if(!section->cert) /* no certificate specified */
+        return 0; /* OK */
 
     ui_data.section=section; /* setup current section for callbacks */
 
@@ -285,7 +321,7 @@ static void load_certificate(LOCAL_OPTIONS *section) {
     if(!SSL_CTX_use_certificate_chain_file(section->ctx, section->cert)) {
         s_log(LOG_ERR, "Error reading certificate file: %s", section->cert);
         sslerror("SSL_CTX_use_certificate_chain_file");
-        die(1);
+        return 1; /* FAILED */
     }
     s_log(LOG_DEBUG, "Certificate loaded");
 
@@ -293,65 +329,55 @@ static void load_certificate(LOCAL_OPTIONS *section) {
     SSL_CTX_set_default_passwd_cb(section->ctx, password_cb);
 #ifdef HAVE_OSSL_ENGINE_H
 #ifdef USE_WIN32
-    uim=UI_create_method("stunnel WIN32 UI");
-    UI_method_set_reader(uim, pin_cb);
-#else
-    uim=NULL;
-#endif
-#endif
-#ifdef HAVE_OSSL_ENGINE_H
+    ui_method=UI_create_method("stunnel WIN32 UI");
+    UI_method_set_reader(ui_method, pin_cb);
+#else /* USE_WIN32 */
+    ui_method=UI_OpenSSL();
+#endif /* USE_WIN32 */
     if(section->engine)
         for(i=1; i<=3; i++) {
             pkey=ENGINE_load_private_key(section->engine, section->key,
-                uim, &ui_data);
+                ui_method, &ui_data);
             if(!pkey) {
                 reason=ERR_GET_REASON(ERR_peek_error());
                 if(i<=2 && (reason==7 || reason==160)) { /* wrong PIN */
-                    sslerror_stack(); /* dump the error stack */
+                    sslerror_queue(); /* dump the error queue */
                     s_log(LOG_ERR, "Wrong PIN: retrying");
                     continue;
                 }
                 sslerror("ENGINE_load_private_key");
-                die(1);
+                return 1; /* FAILED */
             }
             if(SSL_CTX_use_PrivateKey(section->ctx, pkey))
                 break; /* success */
             sslerror("SSL_CTX_use_PrivateKey");
-            die(1);
+            return 1; /* FAILED */
         }
     else
-#endif
+#endif /* HAVE_OSSL_ENGINE_H */
         for(i=0; i<=3; i++) {
             if(!i && !cache_initialized)
                 continue; /* there is no cached value */
             SSL_CTX_set_default_passwd_cb_userdata(section->ctx,
                 i ? &ui_data : NULL); /* try the cached password first */
-#ifdef NO_RSA
             if(SSL_CTX_use_PrivateKey_file(section->ctx, section->key,
                     SSL_FILETYPE_PEM))
-#else /* NO_RSA */
-            if(SSL_CTX_use_RSAPrivateKey_file(section->ctx, section->key,
-                    SSL_FILETYPE_PEM))
-#endif /* NO_RSA */
                 break;
             reason=ERR_GET_REASON(ERR_peek_error());
             if(i<=2 && reason==EVP_R_BAD_DECRYPT) {
-                sslerror_stack(); /* dump the error stack */
+                sslerror_queue(); /* dump the error queue */
                 s_log(LOG_ERR, "Wrong pass phrase: retrying");
                 continue;
             }
-#ifdef NO_RSA
             sslerror("SSL_CTX_use_PrivateKey_file");
-#else /* NO_RSA */
-            sslerror("SSL_CTX_use_RSAPrivateKey_file");
-#endif /* NO_RSA */
-            die(1);
+            return 1; /* FAILED */
         }
     if(!SSL_CTX_check_private_key(section->ctx)) {
         sslerror("Private key does not match the certificate");
-        die(1);
+        return 1; /* FAILED */
     }
     s_log(LOG_DEBUG, "Private key loaded");
+    return 0; /* OK */
 }
 
 static int password_cb(char *buf, int size, int rwflag, void *userdata) {
@@ -390,14 +416,14 @@ static int sess_new_cb(SSL *ssl, SSL_SESSION *sess) {
     int val_len;
 
     val_len=i2d_SSL_SESSION(sess, NULL);
-    val_tmp=val=malloc(val_len);
+    val_tmp=val=str_alloc(val_len);
     if(!val)
         return 1;
     i2d_SSL_SESSION(sess, &val_tmp);
 
     cache_transfer(ssl->ctx, CACHE_CMD_NEW, SSL_SESSION_get_timeout(sess),
         sess->session_id, sess->session_id_length, val, val_len, NULL, NULL);
-    free(val);
+    str_free(val);
     return 1; /* leave the session in local cache for reuse */
 }
 
@@ -414,7 +440,7 @@ static SSL_SESSION *sess_get_cb(SSL *ssl,
         return NULL;
     val_tmp=val;
     sess=d2i_SSL_SESSION(NULL, (const unsigned char **)&val_tmp, val_len);
-    free(val);
+    str_free(val);
     return sess;
 }
 
@@ -439,11 +465,12 @@ static void cache_transfer(SSL_CTX *ctx, const unsigned int type,
     char session_id_txt[2*SSL_MAX_SSL_SESSION_ID_LENGTH+1];
     const char hex[16]="0123456789ABCDEF";
     const char *type_description[]={"new", "get", "remove"};
-    int i, s, len;
+    unsigned int i;
+    int s, len;
     SOCKADDR_UNION addr;
     struct timeval t;
     CACHE_PACKET *packet;
-    LOCAL_OPTIONS *opt;
+    SERVICE_OPTIONS *section;
 
     if(ret) /* set error as the default result if required */
         *ret=NULL;
@@ -469,7 +496,7 @@ static void cache_transfer(SSL_CTX *ctx, const unsigned int type,
             key_len);
         return;
     }
-    packet=calloc(1, sizeof(CACHE_PACKET));
+    packet=str_alloc(sizeof(CACHE_PACKET));
     if(!packet) {
         s_log(LOG_ERR, "cache_transfer: packet buffer allocation failed");
         return;
@@ -478,101 +505,96 @@ static void cache_transfer(SSL_CTX *ctx, const unsigned int type,
     /* setup packet */
     packet->version=1;
     packet->type=type;
-    packet->timeout=htons(timeout<64800?timeout:64800); /* 18 hours */
+    packet->timeout=htons((u_short)(timeout<64800?timeout:64800));/* 18 hours */
     memcpy(packet->key, key, key_len);
     memcpy(packet->val, val, val_len);
 
     /* create the socket */
-    s=socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);
-    if(s==-1) {
-        sockerror("cache_transfer: socket");
-        free(packet);
+    s=s_socket(AF_INET, SOCK_DGRAM, 0, 0, "cache_transfer: socket");
+    if(s<0) {
+        str_free(packet);
         return;
     }
 
     /* retrieve pointer to the section structure of this ctx */
-    opt=SSL_CTX_get_ex_data(ctx, opt_index);
-    memcpy(&addr, &opt->sessiond_addr.addr[0], sizeof addr);
+    section=SSL_CTX_get_ex_data(ctx, opt_index);
+    memcpy(&addr, &section->sessiond_addr.addr[0], sizeof addr);
     if(sendto(s, (void *)packet, sizeof(CACHE_PACKET)-MAX_VAL_LEN+val_len, 0,
-            &addr.sa, addr_len(addr))==-1) {
+            &addr.sa, addr_len(addr))<0) {
         sockerror("cache_transfer: sendto");
         closesocket(s);
-        free(packet);
+        str_free(packet);
         return;
     }
 
     if(!ret || !ret_len) { /* no response is required */
         closesocket(s);
-        free(packet);
+        str_free(packet);
         return;
     }
 
     /* set recvfrom timeout to 200ms */
     t.tv_sec=0;
     t.tv_usec=200;
-    if(setsockopt(s, SOL_SOCKET, SO_RCVTIMEO, (void *)&t, sizeof t)==-1) {
+    if(setsockopt(s, SOL_SOCKET, SO_RCVTIMEO, (void *)&t, sizeof t)<0) {
         sockerror("cache_transfer: setsockopt SO_RCVTIMEO");
         closesocket(s);
-        free(packet);
+        str_free(packet);
         return;
     }
 
     /* retrieve response */
     len=recv(s, (void *)packet, sizeof(CACHE_PACKET), 0);
     closesocket(s);
-    if(len==-1) {
+    if(len<0) {
         if(get_last_socket_error()==EAGAIN)
             s_log(LOG_INFO, "cache_transfer: recv timeout");
         else
             sockerror("cache_transfer: recv");
-        free(packet);
+        str_free(packet);
         return;
     }
 
     /* parse results */
-    if(len<sizeof(CACHE_PACKET)-MAX_VAL_LEN || /* too short */
+    if(len<(int)sizeof(CACHE_PACKET)-MAX_VAL_LEN || /* too short */
             packet->version!=1 || /* wrong version */
             memcmp(packet->key, key, key_len)) { /* wrong session id */
         s_log(LOG_DEBUG, "cache_transfer: malformed packet received");
-        free(packet);
+        str_free(packet);
         return;
     }
     if(packet->type!=CACHE_RESP_OK) {
         s_log(LOG_INFO, "cache_transfer: session not found");
-        free(packet);
+        str_free(packet);
         return;
     }
     *ret_len=len-(sizeof(CACHE_PACKET)-MAX_VAL_LEN);
-    *ret=malloc(*ret_len);
+    *ret=str_alloc(*ret_len);
     if(!*ret) {
         s_log(LOG_ERR, "cache_transfer: return value allocation failed");
-        free(packet);
+        str_free(packet);
         return;
     }
     s_log(LOG_INFO, "cache_transfer: session found");
     memcpy(*ret, packet->val, *ret_len);
-    free(packet);
+    str_free(packet);
 }
 
 /**************************************** informational callback */
 
-#if SSLEAY_VERSION_NUMBER >= 0x00907000L
-static void info_callback(const SSL *s, int where, int ret) {
-#else /* OpenSSL-0.9.7 */
-static void info_callback(SSL *s, int where, int ret) {
-#endif /* OpenSSL-0.9.7 */
+static void info_callback(const SSL *ssl, int where, int ret) {
     if(where & SSL_CB_LOOP)
         s_log(LOG_DEBUG, "SSL state (%s): %s",
         where & SSL_ST_CONNECT ? "connect" :
         where & SSL_ST_ACCEPT ? "accept" :
-        "undefined", SSL_state_string_long(s));
+        "undefined", SSL_state_string_long(ssl));
     else if(where & SSL_CB_ALERT)
         s_log(LOG_DEBUG, "SSL alert (%s): %s: %s",
             where & SSL_CB_READ ? "read" : "write",
             SSL_alert_type_string_long(ret),
             SSL_alert_desc_string_long(ret));
     else if(where==SSL_CB_HANDSHAKE_DONE)
-        print_stats(s->ctx);
+        print_stats(ssl->ctx);
 }
 
 static void print_stats(SSL_CTX *ctx) { /* print statistics */
@@ -582,18 +604,14 @@ static void print_stats(SSL_CTX *ctx) { /* print statistics */
         SSL_CTX_sess_connect(ctx));
     s_log(LOG_DEBUG, "%4ld client connects that finished",
         SSL_CTX_sess_connect_good(ctx));
-#if SSLEAY_VERSION_NUMBER >= 0x0922
     s_log(LOG_DEBUG, "%4ld client renegotiations requested",
         SSL_CTX_sess_connect_renegotiate(ctx));
-#endif
     s_log(LOG_DEBUG, "%4ld server connects (SSL_accept())",
         SSL_CTX_sess_accept(ctx));
     s_log(LOG_DEBUG, "%4ld server connects that finished",
         SSL_CTX_sess_accept_good(ctx));
-#if SSLEAY_VERSION_NUMBER >= 0x0922
     s_log(LOG_DEBUG, "%4ld server renegotiations requested",
         SSL_CTX_sess_accept_renegotiate(ctx));
-#endif
     s_log(LOG_DEBUG, "%4ld session cache hits",
         SSL_CTX_sess_hits(ctx));
     s_log(LOG_DEBUG, "%4ld external session cache hits",
@@ -606,30 +624,35 @@ static void print_stats(SSL_CTX *ctx) { /* print statistics */
 
 /**************************************** SSL error reporting */
 
-void sslerror(char *txt) { /* SSL Error handler */
+void sslerror(char *txt) { /* OpenSSL error handler */
     unsigned long err;
-    char string[120];
 
     err=ERR_get_error();
-    if(!err) {
+    if(err) {
+        sslerror_queue();
+        sslerror_log(err, txt);
+    } else {
         s_log(LOG_ERR, "%s: Peer suddenly disconnected", txt);
-        return;
     }
-    sslerror_stack();
-    ERR_error_string(err, string);
-    s_log(LOG_ERR, "%s: %lX: %s", txt, err, string);
 }
 
-static void sslerror_stack(void) { /* recursive dump of the error stack */
+static void sslerror_queue(void) { /* recursive dump of the error queue */
     unsigned long err;
-    char string[120];
 
     err=ERR_get_error();
-    if(!err)
-        return;
-    sslerror_stack();
-    ERR_error_string(err, string);
-    s_log(LOG_ERR, "error stack: %lX : %s", err, string);
+    if(err) {
+        sslerror_queue();
+        sslerror_log(err, "error queue");
+    }
 }
 
-/* End of ctx.c */
+static void sslerror_log(unsigned long err, char *txt) {
+    char *error_string;
+
+    error_string=str_alloc(120);
+    ERR_error_string(err, error_string);
+    s_log(LOG_ERR, "%s: %lX: %s", txt, err, error_string);
+    str_free(error_string);
+}
+
+/* end of ctx.c */
