@@ -1,6 +1,6 @@
 /*
  *   stunnel       Universal SSL tunnel
- *   Copyright (C) 1998-2009 Michal Trojnara <Michal.Trojnara@mirt.net>
+ *   Copyright (C) 1998-2011 Michal Trojnara <Michal.Trojnara@mirt.net>
  *
  *   This program is free software; you can redistribute it and/or modify it
  *   under the terms of the GNU General Public License as published by the
@@ -40,10 +40,7 @@
 
 /* #define DEBUG_UCONTEXT */
 
-#ifndef USE_WIN32
 static int signal_pipe[2]={-1, -1};
-static char signal_buffer[16];
-static void sigchld_handler(int);
 #ifdef __INNOTEK_LIBC__
 struct sockaddr_un {
     u_char  sun_len;             /* sockaddr len including null */
@@ -55,7 +52,6 @@ static void signal_pipe_empty(void);
 #ifdef USE_FORK
 static void client_status(void); /* dead children detected */
 #endif
-#endif /* !defined(USE_WIN32) */
 
 /**************************************** s_poll functions */
 
@@ -66,7 +62,7 @@ void s_poll_init(s_poll_set *fds) {
 }
 
 void s_poll_add(s_poll_set *fds, int fd, int rd, int wr) {
-    int i;
+    unsigned int i;
 
     for(i=0; i<fds->nfds && fds->ufds[i].fd!=fd; i++)
         ;
@@ -87,20 +83,30 @@ void s_poll_add(s_poll_set *fds, int fd, int rd, int wr) {
 }
 
 int s_poll_canread(s_poll_set *fds, int fd) {
-    int i;
+    unsigned int i;
 
     for(i=0; i<fds->nfds; i++)
         if(fds->ufds[i].fd==fd)
-            return fds->ufds[i].revents&~POLLOUT; /* read or error */
+            return fds->ufds[i].revents&(POLLIN|POLLHUP); /* read or closed */
     return 0;
 }
 
 int s_poll_canwrite(s_poll_set *fds, int fd) {
-    int i;
+    unsigned int i;
 
     for(i=0; i<fds->nfds; i++)
         if(fds->ufds[i].fd==fd)
-            return fds->ufds[i].revents&POLLOUT; /* write */
+            return fds->ufds[i].revents&POLLOUT; /* it is possible to write */
+    return 0;
+}
+
+int s_poll_error(s_poll_set *fds, int fd) {
+    unsigned int i;
+
+    for(i=0; i<fds->nfds; i++)
+        if(fds->ufds[i].fd==fd)
+            return fds->ufds[i].revents&(POLLERR|POLLNVAL) ?
+                get_socket_error(fd) : 0;
     return 0;
 }
 
@@ -109,27 +115,27 @@ int s_poll_canwrite(s_poll_set *fds, int fd) {
 /* move ready contexts from waiting queue to ready queue */
 static void scan_waiting_queue(void) {
     int retval, retry;
-    CONTEXT *ctx, *prev;
+    CONTEXT *context, *prev;
     int min_timeout;
-    int nfds, i;
+    unsigned int nfds, i;
     time_t now;
     short *signal_revents;
-    static int max_nfds=0;
+    static unsigned int max_nfds=0;
     static struct pollfd *ufds=NULL;
 
     time(&now);
     /* count file descriptors */
     min_timeout=-1;
     nfds=0;
-    for(ctx=waiting_head; ctx; ctx=ctx->next) {
-        nfds+=ctx->fds->nfds;
-        if(ctx->finish>=0) /* finite time */
-            if(min_timeout<0 || min_timeout>ctx->finish-now)
-                min_timeout=ctx->finish-now<0 ? 0 : ctx->finish-now;
+    for(context=waiting_head; context; context=context->next) {
+        nfds+=context->fds->nfds;
+        if(context->finish>=0) /* finite time */
+            if(min_timeout<0 || min_timeout>context->finish-now)
+                min_timeout=context->finish-now<0 ? 0 : context->finish-now;
     }
     /* setup ufds structure */
     if(nfds>max_nfds) { /* need to allocate more memory */
-        ufds=realloc(ufds, nfds*sizeof(struct pollfd));
+        ufds=str_realloc(ufds, nfds*sizeof(struct pollfd));
         if(!ufds) {
             s_log(LOG_CRIT, "Memory allocation failed");
             die(1);
@@ -138,10 +144,10 @@ static void scan_waiting_queue(void) {
     }
     nfds=0;
     signal_revents=NULL;
-    for(ctx=waiting_head; ctx; ctx=ctx->next)
-        for(i=0; i<ctx->fds->nfds; i++) {
-            ufds[nfds].fd=ctx->fds->ufds[i].fd;
-            ufds[nfds].events=ctx->fds->ufds[i].events;
+    for(context=waiting_head; context; context=context->next)
+        for(i=0; i<context->fds->nfds; i++) {
+            ufds[nfds].fd=context->fds->ufds[i].fd;
+            ufds[nfds].events=context->fds->ufds[i].events;
             if(ufds[nfds].fd==signal_pipe[0])
                 signal_revents=&ufds[nfds].revents;
             nfds++;
@@ -163,99 +169,104 @@ static void scan_waiting_queue(void) {
     /* process the returned data */
     nfds=0;
     prev=NULL; /* previous element of the waiting queue */
-    ctx=waiting_head;
-    while(ctx) {
-        ctx->ready=0;
+    context=waiting_head;
+    while(context) {
+        context->ready=0;
         /* count ready file descriptors in each context */
-        for(i=0; i<ctx->fds->nfds; i++) {
-            ctx->fds->ufds[i].revents=ufds[nfds].revents;
+        for(i=0; i<context->fds->nfds; i++) {
+            context->fds->ufds[i].revents=ufds[nfds].revents;
 #ifdef DEBUG_UCONTEXT
-            s_log(LOG_DEBUG, "CONTEXT %ld, FD=%d, (%s%s)->(%s%s%s%s%s)",
-                ctx->id, ufds[nfds].fd,
-                ufds[nfds].events & POLLIN ? "IN" : "",
-                ufds[nfds].events & POLLOUT ? "OUT" : "",
-                ufds[nfds].revents & POLLIN ? "IN" : "",
-                ufds[nfds].revents & POLLOUT ? "OUT" : "",
-                ufds[nfds].revents & POLLERR ? "ERR" : "",
-                ufds[nfds].revents & POLLHUP ? "HUP" : "",
-                ufds[nfds].revents & POLLNVAL ? "NVAL" : "");
+            s_log(LOG_DEBUG, "CONTEXT %ld, FD=%d,%s%s ->%s%s%s%s%s",
+                context->id, ufds[nfds].fd,
+                ufds[nfds].events & POLLIN ? " IN" : "",
+                ufds[nfds].events & POLLOUT ? " OUT" : "",
+                ufds[nfds].revents & POLLIN ? " IN" : "",
+                ufds[nfds].revents & POLLOUT ? " OUT" : "",
+                ufds[nfds].revents & POLLERR ? " ERR" : "",
+                ufds[nfds].revents & POLLHUP ? " HUP" : "",
+                ufds[nfds].revents & POLLNVAL ? " NVAL" : "");
 #endif
             if(ufds[nfds].revents)
-                ctx->ready++;
+                context->ready++;
             nfds++;
         }
-        if(ctx->ready || (ctx->finish>=0 && ctx->finish<=now)) {
-            /* remove context ctx from the waiting queue */
+        if(context->ready || (context->finish>=0 && context->finish<=now)) {
+            /* remove context from the waiting queue */
             if(prev)
-                prev->next=ctx->next;
+                prev->next=context->next;
             else
-                waiting_head=ctx->next;
-            if(!ctx->next) /* same as ctx==waiting_tail */
+                waiting_head=context->next;
+            if(!context->next) /* same as context==waiting_tail */
                 waiting_tail=prev;
 
-            /* append context ctx to the ready queue */
-            ctx->next=NULL;
+            /* append context context to the ready queue */
+            context->next=NULL;
             if(ready_tail)
-                ready_tail->next=ctx;
-            ready_tail=ctx;
+                ready_tail->next=context;
+            ready_tail=context;
             if(!ready_head)
-                ready_head=ctx;
-        } else { /* leave the context ctx in the waiting queue */
-            prev=ctx;
+                ready_head=context;
+        } else { /* leave the context context in the waiting queue */
+            prev=context;
         }
-        ctx=prev ? prev->next : waiting_head;
+        context=prev ? prev->next : waiting_head;
     }
 }
 
 int s_poll_wait(s_poll_set *fds, int sec, int msec) {
-    /* FIXME: msec parameter is currently ignored with UCONTEXT threads */
-    CONTEXT *ctx; /* current context */
+    CONTEXT *context; /* current context */
     static CONTEXT *to_free=NULL; /* delayed memory deallocation */
 
+    /* FIXME: msec parameter is currently ignored with UCONTEXT threads */
+    (void)msec; /* skip warning about unused parameter */
+
     /* remove the current context from ready queue */
-    ctx=ready_head;
+    context=ready_head;
     ready_head=ready_head->next;
     if(!ready_head) /* the queue is empty */
         ready_tail=NULL;
+    /* it it safe to s_log() after new ready_head is set */
 
+    /* it's illegal to deallocate the stack of the current context */
+    if(to_free) { /* a delayed deallocation is scheduled */
+        s_log(LOG_DEBUG, "Releasing context %ld", to_free->id);
+        free(to_free->stack);
+        free(to_free);
+        to_free=NULL;
+    }
+
+    /* manage the current thread */
     if(fds) { /* something to wait for -> swap the context */
-        ctx->fds=fds; /* set file descriptors to wait for */
-        ctx->finish=sec<0 ? -1 : time(NULL)+sec;
-        /* move (append) the current context to the waiting queue */
-        ctx->next=NULL;
+        context->fds=fds; /* set file descriptors to wait for */
+        context->finish=sec<0 ? -1 : time(NULL)+sec;
+
+        /* append the current context to the waiting queue */
+        context->next=NULL;
         if(waiting_tail)
-            waiting_tail->next=ctx;
-        waiting_tail=ctx;
+            waiting_tail->next=context;
+        waiting_tail=context;
         if(!waiting_head)
-            waiting_head=ctx;
-        while(!ready_head) /* no context ready */
-            scan_waiting_queue();
-        if(ctx->id!=ready_head->id) {
+            waiting_head=context;
+    } else { /* nothing to wait for -> drop the context */
+        to_free=context; /* schedule for delayed deallocation */
+    }
+
+    while(!ready_head) /* wait until there is a thread to switch to */
+        scan_waiting_queue();
+
+    /* switch threads */
+    if(fds) { /* swap the current context */
+        if(context->id!=ready_head->id) {
             s_log(LOG_DEBUG, "Context swap: %ld -> %ld",
-                ctx->id, ready_head->id);
-            swapcontext(&ctx->ctx, &ready_head->ctx);
+                context->id, ready_head->id);
+            swapcontext(&context->context, &ready_head->context);
             s_log(LOG_DEBUG, "Current context: %ld", ready_head->id);
-            if(to_free) {
-                s_log(LOG_DEBUG, "Releasing context %ld", to_free->id);
-                free(to_free->stack);
-                free(to_free);
-                to_free=NULL;
-            }
         }
         return ready_head->ready;
-    } else { /* nothing to wait for -> drop the context */
-        /* it's illegal to deallocate the stack of the current context */
-        if(to_free) {
-            s_log(LOG_DEBUG, "Releasing context %ld", to_free->id);
-            free(to_free->stack);
-            free(to_free);
-        }
-        to_free=ctx;
-        while(!ready_head) /* no context ready */
-            scan_waiting_queue();
+    } else { /* drop the current context */
         s_log(LOG_DEBUG, "Context set: %ld (dropped) -> %ld",
-            ctx->id, ready_head->id);
-        setcontext(&ready_head->ctx);
+            context->id, ready_head->id);
+        setcontext(&ready_head->context);
         ioerror("setcontext"); /* should not ever happen */
         return 0;
     }
@@ -284,16 +295,16 @@ int s_poll_wait(s_poll_set *fds, int sec, int msec) {
 void s_poll_init(s_poll_set *fds) {
     FD_ZERO(&fds->irfds);
     FD_ZERO(&fds->iwfds);
-    fds->max = 0; /* no file descriptors */
+    fds->max=0; /* no file descriptors */
 }
 
 void s_poll_add(s_poll_set *fds, int fd, int rd, int wr) {
     if(rd)
-        FD_SET(fd, &fds->irfds);
+        FD_SET((unsigned int)fd, &fds->irfds);
     if(wr)
-        FD_SET(fd, &fds->iwfds);
-    if(fd > fds->max)
-        fds->max = fd;
+        FD_SET((unsigned int)fd, &fds->iwfds);
+    if(fd>fds->max)
+        fds->max=fd;
 }
 
 int s_poll_canread(s_poll_set *fds, int fd) {
@@ -304,6 +315,12 @@ int s_poll_canwrite(s_poll_set *fds, int fd) {
     return FD_ISSET(fd, &fds->owfds);
 }
 
+int s_poll_error(s_poll_set *fds, int fd) {
+    if(!FD_ISSET(fd, &fds->orfds)) /* error conditions are signaled as read */
+        return 0;
+    return get_socket_error(fd); /* check if it's really an error */
+}
+
 int s_poll_wait(s_poll_set *fds, int sec, int msec) {
     int retval, retry;
     struct timeval tv, *tv_ptr;
@@ -320,109 +337,120 @@ int s_poll_wait(s_poll_set *fds, int sec, int msec) {
             tv_ptr=&tv;
         }
         retval=select(fds->max+1, &fds->orfds, &fds->owfds, NULL, tv_ptr);
-#if !defined(USE_WIN32) && !defined(USE_OS2)
         if(sec<0 && retval>0 && s_poll_canread(fds, signal_pipe[0])) {
             signal_pipe_empty(); /* no timeout -> main loop */
             retry=1;
         }
-#endif
     } while(retry || (retval<0 && get_last_socket_error()==EINTR));
     return retval;
 }
 
 #endif /* USE_POLL */
 
-#ifndef USE_WIN32
-
-static void sigchld_handler(int sig) { /* SIGCHLD detected */
-    int save_errno;
-#ifdef __sgi
-    int status;
-#endif
-
-    save_errno=errno;
-#ifdef __sgi
-    while(wait_for_pid(-1, &status, WNOHANG)>0) {
-        /* no logging is possible in a signal handler */
-#ifdef USE_FORK
-        --num_clients; /* one client less */
-#endif /* USE_FORK */
-    }
-#else /* __sgi */
-#ifdef __INNOTEK_LIBC__
-    writesocket(signal_pipe[1], signal_buffer, 1);
-#else
-    write(signal_pipe[1], signal_buffer, 1);
-#endif /* __INNOTEK_LIBC__ */
-#endif /* __sgi */
-    signal(SIGCHLD, sigchld_handler);
-    errno=save_errno;
-}
-
-/**************************************** signal pipe */
+/**************************************** signal pipe handling */
 
 int signal_pipe_init(void) {
+#ifdef USE_WIN32
+    if(make_sockets(signal_pipe))
+        die(1);
+#else
 #if defined(__INNOTEK_LIBC__)
-    /* Innotek port of GCC can not use select on a pipe, use local socket instead */
+    /* Innotek port of GCC can not use select on a pipe
+     * use local socket instead */
     struct sockaddr_un un;
     fd_set set_pipe;
     int pipe_in;
 
     FD_ZERO(&set_pipe);
-    signal_pipe[0]=socket(PF_OS2, SOCK_STREAM, 0);
-    pipe_in=signal_pipe[0];
-    signal_pipe[1]=socket(PF_OS2, SOCK_STREAM, 0);
+    signal_pipe[0]=s_socket(PF_OS2, SOCK_STREAM, 0, 0, "socket#1");
+    signal_pipe[1]=s_socket(PF_OS2, SOCK_STREAM, 0, 0, "socket#2");
 
-    alloc_fd(signal_pipe[0]);
-    alloc_fd(signal_pipe[1]);
-
-    /* Connect the two endpoints */
+    /* connect the two endpoints */
     memset(&un, 0, sizeof un);
-
     un.sun_len=sizeof un;
     un.sun_family=AF_OS2;
     sprintf(un.sun_path, "\\socket\\stunnel-%u", getpid());
-    /* Make the first endpoint listen */
+    /* make the first endpoint listen */
     bind(signal_pipe[0], (struct sockaddr *)&un, sizeof un);
-    listen(signal_pipe[0], 5);
+    listen(signal_pipe[0], 1);
     connect(signal_pipe[1], (struct sockaddr *)&un, sizeof un);
     FD_SET(signal_pipe[0], &set_pipe);
-    if (select(signal_pipe[0]+1, &set_pipe, NULL, NULL, NULL)>0) {
-        signal_pipe[0]=accept(signal_pipe[0], NULL, 0);
+    if(select(signal_pipe[0]+1, &set_pipe, NULL, NULL, NULL)>0) {
+        pipe_in=signal_pipe[0];
+        signal_pipe[0]=s_accept(signal_pipe[0], NULL, 0, 0, "accept");
         closesocket(pipe_in);
     } else {
         sockerror("select");
         die(1);
     }
 #else /* __INNOTEK_LIBC__ */
-    if(pipe(signal_pipe)) {
-        ioerror("pipe");
+    if(s_pipe(signal_pipe, 1, "signal_pipe"))
         die(1);
-    }
-    alloc_fd(signal_pipe[0]);
-    alloc_fd(signal_pipe[1]);
-#ifdef FD_CLOEXEC
-    /* close the pipe in child execvp */
-    fcntl(signal_pipe[0], F_SETFD, FD_CLOEXEC);
-    fcntl(signal_pipe[1], F_SETFD, FD_CLOEXEC);
-#endif /* FD_CLOEXEC */
 #endif /* __INNOTEK_LIBC__ */
-    signal(SIGCHLD, sigchld_handler);
+
+    signal(SIGCHLD, signal_handler); /* a child has died */
+    signal(SIGHUP, signal_handler); /* configuration reload */
+    signal(SIGUSR1, signal_handler); /* log reopen */
+    signal(SIGPIPE, SIG_IGN); /* ignore "broken pipe" */
+    if(signal(SIGTERM, SIG_IGN)!=SIG_IGN)
+        signal(SIGTERM, signal_handler); /* fatal */
+    if(signal(SIGQUIT, SIG_IGN)!=SIG_IGN)
+        signal(SIGQUIT, signal_handler); /* fatal */
+    if(signal(SIGINT, SIG_IGN)!=SIG_IGN)
+        signal(SIGINT, signal_handler); /* fatal */
+    /* signal(SIGSEGV, signal_handler); */
+#endif /* USE_WIN32 */
     return signal_pipe[0];
 }
 
+void signal_post(int sig) {
+    writesocket(signal_pipe[1], (char *)&sig, sizeof sig);
+}
+
 static void signal_pipe_empty(void) {
-    s_log(LOG_DEBUG, "Cleaning up the signal pipe");
-#ifdef __INNOTEK_LIBC__
-    readsocket(signal_pipe[0], signal_buffer, sizeof signal_buffer);
-#else
-    read(signal_pipe[0], signal_buffer, sizeof signal_buffer);
-#endif
+    int sig, err;
+
+    s_log(LOG_DEBUG, "Dispatching signals from the signal pipe");
+    while(readsocket(signal_pipe[0], (char *)&sig, sizeof sig)==sizeof sig) {
+        switch(sig) {
+#ifndef USE_WIN32
+        case SIGCHLD:
+            s_log(LOG_DEBUG, "Processing SIGCHLD");
 #ifdef USE_FORK
-    client_status(); /* report status of client process */
+            client_status(); /* report status of client process */
 #else /* USE_UCONTEXT || USE_PTHREAD */
-    child_status();  /* report status of libwrap or 'exec' process */
+            child_status();  /* report status of libwrap or 'exec' process */
 #endif /* defined USE_FORK */
+            break;
+#endif /* !defind USE_WIN32 */
+        case SIGNAL_RELOAD_CONFIG:
+            s_log(LOG_DEBUG, "Processing SIGNAL_RELOAD_CONFIG");
+            unbind_ports();
+            log_close();
+            err=parse_conf(NULL, CONF_RELOAD);
+            if(err) {
+                log_flush(LOG_MODE_ERROR);
+            } else {
+                log_open();
+                bind_ports();
+            }
+#ifdef USE_WIN32
+            win_newconfig(err);
+#endif
+            break;
+        case SIGNAL_REOPEN_LOG:
+            s_log(LOG_DEBUG, "Processing SIGNAL_REOPEN_LOG");
+            log_close();
+            log_open();
+            s_log(LOG_NOTICE, "Log file reopened");
+            break;
+        default:
+            s_log(sig==SIGNAL_TERMINATE ? LOG_NOTICE : LOG_ERR,
+                "Received signal %d; terminating", sig);
+            die(3);
+        }
+    }
+    s_log(LOG_DEBUG, "Signal pipe is empty");
 }
 
 #ifdef USE_FORK
@@ -453,6 +481,8 @@ static void client_status(void) { /* dead children detected */
 }
 #endif /* defined USE_FORK */
 
+#if !defined(USE_WIN32) && !defined(USE_OS2)
+
 void child_status(void) { /* dead libwrap or 'exec' process detected */
     int pid, status;
 
@@ -476,48 +506,19 @@ void child_status(void) { /* dead libwrap or 'exec' process detected */
     }
 }
 
-#endif /* !defined USE_WIN32 */
+void signal_handler(int sig) {
+    int saved_errno;
+
+    saved_errno=errno;
+    signal_post(sig);
+    signal(sig, signal_handler);
+    errno=saved_errno;
+}
+
+#endif /* !defined(USE_WIN32) && !defined(USE_OS2) */
 
 /**************************************** fd management */
 
-int alloc_fd(int sock) {
-#ifndef USE_WIN32
-    if(!max_fds || sock>=max_fds) {
-        s_log(LOG_ERR,
-            "File descriptor out of range (%d>=%d)", sock, max_fds);
-        closesocket(sock);
-        return -1;
-    }
-#endif
-    setnonblock(sock, 1);
-    return 0;
-}
-
-/* Try to use non-POSIX O_NDELAY on obsolete BSD systems */
-#if !defined O_NONBLOCK && defined O_NDELAY
-#define O_NONBLOCK O_NDELAY
-#endif
-
-void setnonblock(int sock, unsigned long l) {
-#if defined F_GETFL && defined F_SETFL && defined O_NONBLOCK && !defined __INNOTEK_LIBC__
-    int retval, flags;
-    do {
-        flags=fcntl(sock, F_GETFL, 0);
-    }while(flags<0 && get_last_socket_error()==EINTR);
-    flags=l ? flags|O_NONBLOCK : flags&~O_NONBLOCK;
-    do {
-        retval=fcntl(sock, F_SETFL, flags);
-    }while(retval<0 && get_last_socket_error()==EINTR);
-    if(retval<0)
-#else
-    if(ioctlsocket(sock, FIONBIO, &l)<0)
-#endif
-        sockerror("nonblocking"); /* non-critical */
-    else
-        s_log(LOG_DEBUG, "FD %d in %sblocking mode", sock,
-            l ? "non-" : "");
-}
-
 int set_socket_options(int s, int type) {
     SOCK_OPT *ptr;
     extern SOCK_OPT sock_opts[];
@@ -529,31 +530,48 @@ int set_socket_options(int s, int type) {
             continue; /* default */
         switch(ptr->opt_type) {
         case TYPE_LINGER:
-            opt_size=sizeof(struct linger); break;
+            opt_size=sizeof(struct linger);
+            break;
         case TYPE_TIMEVAL:
-            opt_size=sizeof(struct timeval); break;
+            opt_size=sizeof(struct timeval);
+            break;
         case TYPE_STRING:
-            opt_size=strlen(ptr->opt_val[type]->c_val)+1; break;
+            opt_size=strlen(ptr->opt_val[type]->c_val)+1;
+            break;
         default:
-            opt_size=sizeof(int); break;
+            opt_size=sizeof(int);
         }
         if(setsockopt(s, ptr->opt_level, ptr->opt_name,
                 (void *)ptr->opt_val[type], opt_size)) {
-            sockerror(ptr->opt_str);
-            return -1; /* FAILED */
+            if(get_last_socket_error()==EOPNOTSUPP) {
+                s_log(LOG_NOTICE,
+                    "Failed to set option %s set on %s socket (not a socket?)",
+                    ptr->opt_str, type_str[type]);
+            } else {
+                sockerror(ptr->opt_str);
+                return -1; /* FAILED */
+            }
         } else {
-            s_log(LOG_DEBUG, "%s option set on %s socket",
+            s_log(LOG_DEBUG, "Option %s set on %s socket",
                 ptr->opt_str, type_str[type]);
         }
     }
     return 0; /* OK */
 }
 
+int get_socket_error(const int fd) {
+    int err;
+    socklen_t optlen=sizeof err;
+
+    if(getsockopt(fd, SOL_SOCKET, SO_ERROR, (void *)&err, &optlen))
+        return get_last_socket_error(); /* failed -> ask why */
+    return err;
+}
+
 /**************************************** simulate blocking I/O */
 
 int connect_blocking(CLI *c, SOCKADDR_UNION *addr, socklen_t addrlen) {
     int error;
-    socklen_t optlen;
     char dst[IPLEN];
 
     s_ntop(dst, addr);
@@ -566,7 +584,7 @@ int connect_blocking(CLI *c, SOCKADDR_UNION *addr, socklen_t addrlen) {
     error=get_last_socket_error();
     if(error!=EINPROGRESS && error!=EWOULDBLOCK) {
         s_log(LOG_ERR, "connect_blocking: connect %s: %s (%d)",
-            dst, my_strerror(error), error);
+            dst, s_strerror(error), error);
         return -1;
     }
 
@@ -578,22 +596,20 @@ int connect_blocking(CLI *c, SOCKADDR_UNION *addr, socklen_t addrlen) {
     case -1:
         error=get_last_socket_error();
         s_log(LOG_ERR, "connect_blocking: s_poll_wait %s: %s (%d)",
-            dst, my_strerror(error), error);
+            dst, s_strerror(error), error);
         return -1;
     case 0:
-        s_log(LOG_ERR, "connect_blocking: s_poll_wait %s: timeout", dst);
+        s_log(LOG_ERR, "connect_blocking: s_poll_wait %s:"
+            " TIMEOUTconnect exceeded", dst);
         return -1;
     default:
-        if(s_poll_canread(&c->fds, c->fd)) {
+        if(s_poll_canread(&c->fds, c->fd) || s_poll_error(&c->fds, c->fd)) {
             /* newly connected socket should not be ready for read */
             /* get the resulting error code, now */
-            optlen=sizeof error;
-            if(getsockopt(c->fd, SOL_SOCKET, SO_ERROR,
-                    (void *)&error, &optlen))
-                error=get_last_socket_error(); /* failed -> ask why */
+            error=get_socket_error(c->fd);
             if(error) { /* really an error? */
                 s_log(LOG_ERR, "connect_blocking: getsockopt %s: %s (%d)",
-                    dst, my_strerror(error), error);
+                    dst, s_strerror(error), error);
                 return -1;
             }
         }
@@ -621,12 +637,13 @@ void write_blocking(CLI *c, int fd, void *ptr, int len) {
             sockerror("write_blocking: s_poll_wait");
             longjmp(c->err, 1); /* error */
         case 0:
-            s_log(LOG_INFO, "write_blocking: s_poll_wait timeout");
+            s_log(LOG_INFO, "write_blocking: s_poll_wait:"
+                " TIMEOUTbusy exceeded: sending reset");
             longjmp(c->err, 1); /* timeout */
         case 1:
             break; /* OK */
         default:
-            s_log(LOG_ERR, "write_blocking: s_poll_wait unknown result");
+            s_log(LOG_ERR, "write_blocking: s_poll_wait: unknown result");
             longjmp(c->err, 1); /* error */
         }
         num=writesocket(fd, ptr, len);
@@ -653,12 +670,13 @@ void read_blocking(CLI *c, int fd, void *ptr, int len) {
             sockerror("read_blocking: s_poll_wait");
             longjmp(c->err, 1); /* error */
         case 0:
-            s_log(LOG_INFO, "read_blocking: s_poll_wait timeout");
+            s_log(LOG_INFO, "read_blocking: s_poll_wait:"
+                " TIMEOUTbusy exceeded: sending reset");
             longjmp(c->err, 1); /* timeout */
         case 1:
             break; /* OK */
         default:
-            s_log(LOG_ERR, "read_blocking: s_poll_wait unknown result");
+            s_log(LOG_ERR, "read_blocking: s_poll_wait: unknown result");
             longjmp(c->err, 1); /* error */
         }
         num=readsocket(fd, ptr, len);
@@ -676,46 +694,57 @@ void read_blocking(CLI *c, int fd, void *ptr, int len) {
 }
 
 void fdputline(CLI *c, int fd, const char *line) {
-    char tmpline[STRLEN];
+    char *tmpline;
     const char crlf[]="\r\n";
     int len;
 
-    safecopy(tmpline, line);
-    safeconcat(tmpline, crlf);
+    tmpline=str_printf("%s%s", line, crlf);
     len=strlen(tmpline);
     write_blocking(c, fd, tmpline, len);
     tmpline[len-2]='\0'; /* remove CRLF */
     safestring(tmpline);
     s_log(LOG_DEBUG, " -> %s", tmpline);
+    str_free(tmpline);
 }
 
-void fdgetline(CLI *c, int fd, char *line) {
-    char tmpline[STRLEN];
+char *fdgetline(CLI *c, int fd) {
+    char *line=NULL, *tmpline;
     s_poll_set fds;
-    int ptr;
+    int ptr=0;
 
-    for(ptr=0;;) {
+    for(;;) {
         s_poll_init(&fds);
         s_poll_add(&fds, fd, 1, 0); /* read */
         switch(s_poll_wait(&fds, c->opt->timeout_busy, 0)) {
         case -1:
             sockerror("fdgetline: s_poll_wait");
+            str_free(line);
             longjmp(c->err, 1); /* error */
         case 0:
-            s_log(LOG_INFO, "fdgetline: s_poll_wait timeout");
+            s_log(LOG_INFO, "fdgetline: s_poll_wait:"
+                " TIMEOUTbusy exceeded: sending reset");
+            str_free(line);
             longjmp(c->err, 1); /* timeout */
         case 1:
             break; /* OK */
         default:
-            s_log(LOG_ERR, "fdgetline: s_poll_wait unknown result");
+            s_log(LOG_ERR, "fdgetline: s_poll_wait: unknown result");
+            str_free(line);
+            longjmp(c->err, 1); /* error */
+        }
+        line=str_realloc(line, ptr+1);
+        if(!line) {
+            s_log(LOG_CRIT, "Memory allocation failed");
             longjmp(c->err, 1); /* error */
         }
         switch(readsocket(fd, line+ptr, 1)) {
         case -1: /* error */
             sockerror("readsocket (fdgetline)");
+            str_free(line);
             longjmp(c->err, 1);
         case 0: /* EOF */
             s_log(LOG_ERR, "Unexpected socket close (fdgetline)");
+            str_free(line);
             longjmp(c->err, 1);
         }
         if(line[ptr]=='\r')
@@ -724,54 +753,94 @@ void fdgetline(CLI *c, int fd, char *line) {
             break;
         if(!line[ptr])
             break;
-        if(++ptr==STRLEN) {
-            s_log(LOG_ERR, "Input line too long");
-            longjmp(c->err, 1);
-        }
+        ++ptr;
     }
     line[ptr]='\0';
-    safecopy(tmpline, line);
+    tmpline=str_dup(line);
     safestring(tmpline);
     s_log(LOG_DEBUG, " <- %s", tmpline);
+    str_free(tmpline);
+    return line;
 }
 
 int fdprintf(CLI *c, int fd, const char *format, ...) {
-    va_list arglist;
-    char line[STRLEN];
-    int len;
+    va_list ap;
+    char *line;
 
-    va_start(arglist, format);
-#ifdef HAVE_VSNPRINTF
-    len=vsnprintf(line, STRLEN-2, format, arglist);
-#else
-    len=vsprintf(line, format, arglist);
-#endif
-    va_end(arglist);
-    if(len<0) {
-        s_log(LOG_ERR, "fdprintf: vs(n)printf failed");
+    va_start(ap, format);
+    line=str_vprintf(format, ap);
+    va_end(ap);
+    if(!line) {
+        s_log(LOG_ERR, "fdprintf: str_vprintf failed");
         longjmp(c->err, 1);
     }
     fdputline(c, fd, line);
-    return len+2;
+    str_free(line);
+    return strlen(line)+2;
 }
 
-int fdscanf(CLI *c, int fd, const char *format, char *buffer) {
-    char line[STRLEN], lformat[STRLEN];
-    int ptr, retval;
+#define INET_SOCKET_PAIR
 
-    fdgetline(c, fd, line);
+int make_sockets(int fd[2]) { /* make a pair of connected ipv4 sockets */
+#ifdef INET_SOCKET_PAIR
+    struct sockaddr_in addr;
+    socklen_t addrlen;
+    int s; /* temporary socket awaiting for connection */
 
-    retval=sscanf(line, format, buffer);
-    if(retval>=0)
-        return retval;
+    /* create two *blocking* sockets first */
+    s=s_socket(AF_INET, SOCK_STREAM, 0, 0, "make_sockets: s_socket#1");
+    if(s<0) {
+        return 1;
+    }
+    fd[1]=s_socket(AF_INET, SOCK_STREAM, 0, 0, "make_sockets: s_socket#2");
+    if(fd[1]<0) {
+        closesocket(s);
+        return 1;
+    }
 
-    s_log(LOG_DEBUG, "fdscanf falling back to lowercase");
-    safecopy(lformat, format);
-    for(ptr=0; lformat[ptr]; ptr++)
-        lformat[ptr]=tolower(lformat[ptr]);
-    for(ptr=0; line[ptr]; ptr++)
-        line[ptr]=tolower(line[ptr]);
-    return sscanf(line, lformat, buffer);
+    addrlen=sizeof addr;
+    memset(&addr, 0, addrlen);
+    addr.sin_family=AF_INET;
+    addr.sin_addr.s_addr=htonl(INADDR_LOOPBACK);
+    addr.sin_port=htons(0); /* dynamic port allocation */
+    if(bind(s, (struct sockaddr *)&addr, addrlen))
+        log_error(LOG_DEBUG, get_last_socket_error(), "make_sockets: bind#1");
+    if(bind(fd[1], (struct sockaddr *)&addr, addrlen))
+        log_error(LOG_DEBUG, get_last_socket_error(), "make_sockets: bind#2");
+
+    if(listen(s, 1)) {
+        sockerror("make_sockets: listen");
+        closesocket(s);
+        closesocket(fd[1]);
+        return 1;
+    }
+    if(getsockname(s, (struct sockaddr *)&addr, &addrlen)) {
+        sockerror("make_sockets: getsockname");
+        closesocket(s);
+        closesocket(fd[1]);
+        return 1;
+    }
+    if(connect(fd[1], (struct sockaddr *)&addr, addrlen)) {
+        sockerror("make_sockets: connect");
+        closesocket(s);
+        closesocket(fd[1]);
+        return 1;
+    }
+    fd[0]=s_accept(s, (struct sockaddr *)&addr, &addrlen, 1,
+        "make_sockets: s_accept");
+    if(fd[0]<0) {
+        closesocket(s);
+        closesocket(fd[1]);
+        return 1;
+    }
+    closesocket(s); /* don't care about the result */
+    set_nonblock(fd[0], 1);
+    set_nonblock(fd[1], 1);
+#else
+    if(s_socketpair(AF_UNIX, SOCK_STREAM, 0, fd, 1, "make_sockets: socketpair"))
+        return 1;
+#endif
+    return 0;
 }
 
-/* End of network.c */
+/* end of network.c */
