@@ -1,6 +1,6 @@
 /*
  *   stunnel       Universal SSL tunnel
- *   Copyright (C) 1998-2009 Michal Trojnara <Michal.Trojnara@mirt.net>
+ *   Copyright (C) 1998-2011 Michal Trojnara <Michal.Trojnara@mirt.net>
  *
  *   This program is free software; you can redistribute it and/or modify it
  *   under the terms of the GNU General Public License as published by the
@@ -48,16 +48,16 @@ int main(int argc, char *argv[]) {
     return 0;
 }
 
-void win_log(LPSTR line) { /* Also used in log.c */
+void win_log(LPSTR line) { /* also used in log.c */
     LPTSTR tstr;
 
     tstr=str2tstr(line);
     RETAILMSG(TRUE, (TEXT("%s\r\n"), tstr));
-    free(tstr);
+    str_free(tstr);
 }
 
-void exit_win32(int code) {
-    exit(code);
+void win_exit(int exit_code) {
+    exit(exit_code);
 }
 
 int passwd_cb(char *buf, int size, int rwflag, void *userdata) {
@@ -70,4 +70,4 @@ int pin_cb(UI *ui, UI_STRING *uis) {
 }
 #endif
 
-/* End of nogui.c */
+/* end of nogui.c */
