@@ -1,6 +1,6 @@
 /*
  *   stunnel       Universal SSL tunnel
- *   Copyright (C) 1998-2009 Michal Trojnara <Michal.Trojnara@mirt.net>
+ *   Copyright (C) 1998-2011 Michal Trojnara <Michal.Trojnara@mirt.net>
  *
  *   This program is free software; you can redistribute it and/or modify it
  *   under the terms of the GNU General Public License as published by the
@@ -41,7 +41,6 @@
 /* \n is not a character expected in the string */
 #define LINE "%[^\n]"
 #define isprefix(a, b) (strncasecmp((a), (b), strlen(b))==0)
-#define s_min(a, b) ((a)>(b)?(b):(a))
 
 /* protocol-specific function prototypes */
 static void cifs_client(CLI *);
@@ -56,14 +55,17 @@ static void imap_client(CLI *);
 static void imap_server(CLI *);
 static void nntp_client(CLI *);
 static void connect_client(CLI *);
+static void ntlm(CLI *);
+#ifndef OPENSSL_NO_MD4
 static char *ntlm1();
 static char *ntlm3(char *, char *, char *);
 static void crypt_DES(DES_cblock, DES_cblock, DES_cblock);
+#endif
 static char *base64(int, char *, int);
 
 void negotiate(CLI *c) {
     if(!c->opt->protocol)
-        return; /* No protocol negotiations */
+        return; /* no protocol negotiations */
 
     s_log(LOG_NOTICE, "Negotiations for %s (%s side) started", c->opt->protocol,
         c->opt->option.client ? "client" : "server");
@@ -150,7 +152,7 @@ static void cifs_server(CLI *c) {
     write_blocking(c, c->local_wfd.fd, response_use_ssl, 5);
 }
 
-/* http://www.postgresql.org/docs/8.3/static/protocol-flow.html#AEN73013 */
+/* http://www.postgresql.org/docs/8.3/static/protocol-flow.html#AEN73982 */
 u8 ssl_request[8]={0, 0, 0, 8, 0x04, 0xd2, 0x16, 0x2f};
 
 static void pgsql_client(CLI *c) {
@@ -179,34 +181,34 @@ static void pgsql_server(CLI *c) {
 }
 
 static void smtp_client(CLI *c) {
-    char line[STRLEN];
+    char *line;
 
-    do { /* Copy multiline greeting */
-        fdgetline(c, c->remote_fd.fd, line);
+    do { /* copy multiline greeting */
+        line=fdgetline(c, c->remote_fd.fd);
         fdputline(c, c->local_wfd.fd, line);
     } while(isprefix(line, "220-"));
 
     fdputline(c, c->remote_fd.fd, "EHLO localhost");
-    do { /* Skip multiline reply */
-        fdgetline(c, c->remote_fd.fd, line);
+    do { /* skip multiline reply */
+        line=fdgetline(c, c->remote_fd.fd);
     } while(isprefix(line, "250-"));
-    if(!isprefix(line, "250 ")) { /* Error */
+    if(!isprefix(line, "250 ")) { /* error */
         s_log(LOG_ERR, "Remote server is not RFC 1425 compliant");
         longjmp(c->err, 1);
     }
 
     fdputline(c, c->remote_fd.fd, "STARTTLS");
-    do { /* Skip multiline reply */
-        fdgetline(c, c->remote_fd.fd, line);
+    do { /* skip multiline reply */
+        line=fdgetline(c, c->remote_fd.fd);
     } while(isprefix(line, "220-"));
-    if(!isprefix(line, "220 ")) { /* Error */
+    if(!isprefix(line, "220 ")) { /* error */
         s_log(LOG_ERR, "Remote server is not RFC 2487 compliant");
         longjmp(c->err, 1);
     }
 }
 
 static void smtp_server(CLI *c) {
-    char line[STRLEN];
+    char *line;
 
     s_poll_init(&c->fds);
     s_poll_add(&c->fds, c->local_rfd.fd, 1, 0);
@@ -216,26 +218,26 @@ static void smtp_server(CLI *c) {
         break;
     case 1: /* fd ready to read */
         s_log(LOG_DEBUG, "RFC 2487 not detected");
-        return; /* Return if RFC 2487 is not used */
+        return; /* return if RFC 2487 is not used */
     default: /* -1 */
         sockerror("RFC2487 (s_poll_wait)");
         longjmp(c->err, 1);
     }
 
-    fdgetline(c, c->remote_fd.fd, line);
+    line=fdgetline(c, c->remote_fd.fd);
     if(!isprefix(line, "220")) {
         s_log(LOG_ERR, "Unknown server welcome");
         longjmp(c->err, 1);
     }
     fdprintf(c, c->local_wfd.fd, "%s + stunnel", line);
-    fdgetline(c, c->local_rfd.fd, line);
+    line=fdgetline(c, c->local_rfd.fd);
     if(!isprefix(line, "EHLO ")) {
         s_log(LOG_ERR, "Unknown client EHLO");
         longjmp(c->err, 1);
     }
     fdprintf(c, c->local_wfd.fd, "250-%s Welcome", line);
     fdputline(c, c->local_wfd.fd, "250 STARTTLS");
-    fdgetline(c, c->local_rfd.fd, line);
+    line=fdgetline(c, c->local_rfd.fd);
     if(!isprefix(line, "STARTTLS")) {
         s_log(LOG_ERR, "STARTTLS expected");
         longjmp(c->err, 1);
@@ -244,16 +246,16 @@ static void smtp_server(CLI *c) {
 }
 
 static void pop3_client(CLI *c) {
-    char line[STRLEN];
+    char *line;
 
-    fdgetline(c, c->remote_fd.fd, line);
+    line=fdgetline(c, c->remote_fd.fd);
     if(!isprefix(line, "+OK ")) {
         s_log(LOG_ERR, "Unknown server welcome");
         longjmp(c->err, 1);
     }
     fdputline(c, c->local_wfd.fd, line);
     fdputline(c, c->remote_fd.fd, "STLS");
-    fdgetline(c, c->remote_fd.fd, line);
+    line=fdgetline(c, c->remote_fd.fd);
     if(!isprefix(line, "+OK ")) {
         s_log(LOG_ERR, "Server does not support TLS");
         longjmp(c->err, 1);
@@ -261,14 +263,14 @@ static void pop3_client(CLI *c) {
 }
 
 static void pop3_server(CLI *c) {
-    char line[STRLEN];
+    char *line;
 
-    fdgetline(c, c->remote_fd.fd, line);
+    line=fdgetline(c, c->remote_fd.fd);
     fdprintf(c, c->local_wfd.fd, "%s + stunnel", line);
-    fdgetline(c, c->local_rfd.fd, line);
-    if(isprefix(line, "CAPA")) { /* Client wants RFC 2449 extensions */
+    line=fdgetline(c, c->local_rfd.fd);
+    if(isprefix(line, "CAPA")) { /* client wants RFC 2449 extensions */
         fdputline(c, c->local_wfd.fd, "-ERR Stunnel does not support capabilities");
-        fdgetline(c, c->local_rfd.fd, line);
+        line=fdgetline(c, c->local_rfd.fd);
     }
     if(!isprefix(line, "STLS")) {
         s_log(LOG_ERR, "Client does not want TLS");
@@ -278,16 +280,16 @@ static void pop3_server(CLI *c) {
 }
 
 static void imap_client(CLI *c) {
-    char line[STRLEN];
+    char *line;
 
-    fdgetline(c, c->remote_fd.fd, line);
+    line=fdgetline(c, c->remote_fd.fd);
     if(!isprefix(line, "* OK")) {
         s_log(LOG_ERR, "Unknown server welcome");
         longjmp(c->err, 1);
     }
     fdputline(c, c->local_wfd.fd, line);
     fdputline(c, c->remote_fd.fd, "stunnel STARTTLS");
-    fdgetline(c, c->remote_fd.fd, line);
+    line=fdgetline(c, c->remote_fd.fd);
     if(!isprefix(line, "stunnel OK")) {
         fdputline(c, c->local_wfd.fd,
             "* BYE stunnel: Server does not support TLS");
@@ -297,7 +299,7 @@ static void imap_client(CLI *c) {
 }
 
 static void imap_server(CLI *c) {
-    char line[STRLEN], id[STRLEN], *tail, *capa;
+    char *line, *id, *tail, *capa;
  
     s_poll_init(&c->fds);
     s_poll_add(&c->fds, c->local_rfd.fd, 1, 0);
@@ -307,14 +309,14 @@ static void imap_server(CLI *c) {
         break;
     case 1: /* fd ready to read */
         s_log(LOG_DEBUG, "RFC 2595 not detected");
-        return; /* Return if RFC 2595 is not used */
+        return; /* return if RFC 2595 is not used */
     default: /* -1 */
         sockerror("RFC2595 (s_poll_wait)");
         longjmp(c->err, 1);
     }
 
     /* process server welcome and send it to client */
-    fdgetline(c, c->remote_fd.fd, line);
+    line=fdgetline(c, c->remote_fd.fd);
     if(!isprefix(line, "* OK")) {
         s_log(LOG_ERR, "Unknown server welcome");
         longjmp(c->err, 1);
@@ -327,9 +329,9 @@ static void imap_server(CLI *c) {
     fdprintf(c, c->local_wfd.fd, "%s (stunnel)", line);
 
     while(1) { /* process client commands */
-        fdgetline(c, c->local_rfd.fd, line);
+        line=fdgetline(c, c->local_rfd.fd);
         /* split line into id and tail */
-        safecopy(id, line);
+        id=str_dup(line);
         tail=strchr(id, ' ');
         if(!tail)
             break;
@@ -341,7 +343,7 @@ static void imap_server(CLI *c) {
             return; /* success */
         } else if(isprefix(tail, "CAPABILITY")) {
             fdputline(c, c->remote_fd.fd, line); /* send it to server */
-            fdgetline(c, c->remote_fd.fd, line); /* get the capabilites */
+            line=fdgetline(c, c->remote_fd.fd); /* get the capabilites */
             if(*line=='*') {
                 /* 
                  * append STARTTLS
@@ -350,7 +352,7 @@ static void imap_server(CLI *c) {
                  * LOGIN would fail as "unexpected command", anyway
                  */
                 fdprintf(c, c->local_wfd.fd, "%s STARTTLS", line);
-                fdgetline(c, c->remote_fd.fd, line); /* next line */
+                line=fdgetline(c, c->remote_fd.fd); /* next line */
             }
             fdputline(c, c->local_wfd.fd, line); /* forward to the client */
             tail=strchr(line, ' ');
@@ -373,23 +375,23 @@ static void imap_server(CLI *c) {
     }    
     /* clean server shutdown */
     fdputline(c, c->remote_fd.fd, "stunnel LOGOUT");
-    fdgetline(c, c->remote_fd.fd, line);
+    line=fdgetline(c, c->remote_fd.fd);
     if(*line=='*')
-        fdgetline(c, c->remote_fd.fd, line);
+        line=fdgetline(c, c->remote_fd.fd);
     longjmp(c->err, 2); /* don't reset */
 }
 
 static void nntp_client(CLI *c) {
-    char line[STRLEN];
+    char *line;
 
-    fdgetline(c, c->remote_fd.fd, line);
+    line=fdgetline(c, c->remote_fd.fd);
     if(!isprefix(line, "200 ") && !isprefix(line, "201 ")) {
         s_log(LOG_ERR, "Unknown server welcome");
         longjmp(c->err, 1);
     }
     fdputline(c, c->local_wfd.fd, line);
     fdputline(c, c->remote_fd.fd, "STARTTLS");
-    fdgetline(c, c->remote_fd.fd, line);
+    line=fdgetline(c, c->remote_fd.fd);
     if(!isprefix(line, "382 ")) {
         s_log(LOG_ERR, "Server does not support TLS");
         longjmp(c->err, 1);
@@ -397,9 +399,7 @@ static void nntp_client(CLI *c) {
 }
 
 static void connect_client(CLI *c) {
-    char line[STRLEN], ntlm2[STRLEN], *encoded;
-    long content_length;
-    char buf[BUFSIZ];
+    char *line, *encoded;
 
     if(!c->opt->protocol_host) {
         s_log(LOG_ERR, "protocolHost not specified");
@@ -410,71 +410,34 @@ static void connect_client(CLI *c) {
     fdprintf(c, c->remote_fd.fd, "Host: %s", c->opt->protocol_host);
     if(c->opt->protocol_username && c->opt->protocol_password) {
         if(!strcasecmp(c->opt->protocol_authentication, "NTLM")) {
-
-            /* send Proxy-Authorization (phase 1) */
-            fdprintf(c, c->remote_fd.fd, "Proxy-Connection: keep-alive");
-            fdprintf(c, c->remote_fd.fd, "Proxy-Authorization: NTLM %s",
-                ntlm1());
-            fdputline(c, c->remote_fd.fd, ""); /* empty line */
-            fdgetline(c, c->remote_fd.fd, line);
-
-            /* receive Proxy-Authenticate (phase 2) */
-            if(line[9]!='4' || line[10]!='0' || line[11]!='7') { /* code 407 */
-                s_log(LOG_ERR, "NTLM authorization request rejected");
-                do { /* read all headers */
-                    fdgetline(c, c->remote_fd.fd, line);
-                } while(*line);
+            ntlm(c);
+        } else { /* basic authentication */
+            line=str_printf("%s:%s",
+                c->opt->protocol_username, c->opt->protocol_password);
+            encoded=base64(1, line, strlen(line));
+            str_free(line);
+            if(!encoded) {
+                s_log(LOG_ERR, "Base64 encoder failed");
                 longjmp(c->err, 1);
             }
-            *ntlm2='\0';
-            content_length=0; /* no HTTP content */
-            do { /* read all headers */
-                fdgetline(c, c->remote_fd.fd, line);
-                if(isprefix(line, "Proxy-Authenticate: NTLM "))
-                    safecopy(ntlm2, line+25);
-                else if(isprefix(line, "Content-Length: "))
-                    content_length=atol(line+16);
-            } while(*line);
-
-            /* read and ignore HTTP content (if any) */
-            while(content_length) {
-                read_blocking(c, c->remote_fd.fd, buf,
-                    s_min(content_length, BUFSIZ));
-                content_length-=s_min(content_length, BUFSIZ);
-            }
-
-            /* send Proxy-Authorization (phase 3) */
-            fdprintf(c, c->remote_fd.fd, "CONNECT %s HTTP/1.1",
-                c->opt->protocol_host);
-            fdprintf(c, c->remote_fd.fd, "Host: %s", c->opt->protocol_host);
-            encoded=ntlm3(c->opt->protocol_username, c->opt->protocol_password,
-                ntlm2);
-            safecopy(line, encoded);
-            free(encoded);
-            fdprintf(c, c->remote_fd.fd, "Proxy-Authorization: NTLM %s", line);
-        } else { /* basic authentication */
-            safecopy(line, c->opt->protocol_username);
-            safeconcat(line, ":");
-            safeconcat(line, c->opt->protocol_password);
-            encoded=base64(1, line, strlen(line));
-            safecopy(line, encoded);
-            free(encoded);
             fdprintf(c, c->remote_fd.fd, "Proxy-Authorization: basic %s",
-                line);
+                encoded);
+            str_free(encoded);
         }
     }
     fdputline(c, c->remote_fd.fd, ""); /* empty line */
-    fdgetline(c, c->remote_fd.fd, line);
-    if(line[9]!='2') { /* "HTTP/1.0 200 Connection established" */
+    line=fdgetline(c, c->remote_fd.fd);
+    if(strlen(line)<12 || line[9]!='2') {
+        /* not "HTTP/1.0 200 Connection established" */
         s_log(LOG_ERR, "CONNECT request rejected");
         do { /* read all headers */
-            fdgetline(c, c->remote_fd.fd, line);
+            line=fdgetline(c, c->remote_fd.fd);
         } while(*line);
         longjmp(c->err, 1);
     }
     s_log(LOG_INFO, "CONNECT request accepted");
     do {
-        fdgetline(c, c->remote_fd.fd, line); /* read all headers */
+        line=fdgetline(c, c->remote_fd.fd); /* read all headers */
     } while(*line);
 }
 
@@ -484,6 +447,71 @@ static void connect_client(CLI *c) {
  * http://www.innovation.ch/personal/ronald/ntlm.html
  */
 
+#define s_min(a, b) ((a)>(b)?(b):(a))
+
+static void ntlm(CLI *c) {
+#ifndef OPENSSL_NO_MD4
+    char *line, buf[BUFSIZ], *ntlm1_txt, *ntlm2_txt, *ntlm3_txt;
+    long content_length=0; /* no HTTP content */
+
+    /* send Proxy-Authorization (phase 1) */
+    fdprintf(c, c->remote_fd.fd, "Proxy-Connection: keep-alive");
+    ntlm1_txt=ntlm1();
+    if(!ntlm1_txt) {
+        s_log(LOG_ERR, "Proxy-Authenticate: Failed to build NTLM request");
+        longjmp(c->err, 1);
+    }
+    fdprintf(c, c->remote_fd.fd, "Proxy-Authorization: NTLM %s", ntlm1_txt);
+    str_free(ntlm1_txt);
+    fdputline(c, c->remote_fd.fd, ""); /* empty line */
+    line=fdgetline(c, c->remote_fd.fd);
+
+    /* receive Proxy-Authenticate (phase 2) */
+    if(line[9]!='4' || line[10]!='0' || line[11]!='7') { /* code 407 */
+        s_log(LOG_ERR, "NTLM authorization request rejected");
+        do { /* read all headers */
+            line=fdgetline(c, c->remote_fd.fd);
+        } while(*line);
+        longjmp(c->err, 1);
+    }
+    ntlm2_txt=NULL;
+    do { /* read all headers */
+        line=fdgetline(c, c->remote_fd.fd);
+        if(isprefix(line, "Proxy-Authenticate: NTLM "))
+            ntlm2_txt=str_dup(line+25);
+        else if(isprefix(line, "Content-Length: "))
+            content_length=atol(line+16);
+    } while(*line);
+    if(!ntlm2_txt) { /* no Proxy-Authenticate: NTLM header */
+        s_log(LOG_ERR, "Proxy-Authenticate: NTLM header not found");
+        longjmp(c->err, 1);
+    }
+
+    /* read and ignore HTTP content (if any) */
+    while(content_length) {
+        read_blocking(c, c->remote_fd.fd, buf, s_min(content_length, BUFSIZ));
+        content_length-=s_min(content_length, BUFSIZ);
+    }
+
+    /* send Proxy-Authorization (phase 3) */
+    fdprintf(c, c->remote_fd.fd, "CONNECT %s HTTP/1.1", c->opt->protocol_host);
+    fdprintf(c, c->remote_fd.fd, "Host: %s", c->opt->protocol_host);
+    ntlm3_txt=ntlm3(c->opt->protocol_username, c->opt->protocol_password, ntlm2_txt);
+    str_free(ntlm2_txt);
+    if(!ntlm3_txt) {
+        s_log(LOG_ERR, "Proxy-Authenticate: Failed to build NTLM response");
+        longjmp(c->err, 1);
+    }
+    fdprintf(c, c->remote_fd.fd, "Proxy-Authorization: NTLM %s", ntlm3_txt);
+    str_free(ntlm3_txt);
+#else
+    s_log(LOG_ERR, "NTLM authentication is not available");
+    longjmp(c->err, 1);
+#endif
+}
+
+#ifndef OPENSSL_NO_MD4
+
 static char *ntlm1() {
     char phase1[16];
 
@@ -500,8 +528,8 @@ static char *ntlm3(char *username, char *password, char *phase2) {
     char *decoded; /* decoded reply from proxy */
     char phase3[146];
     unsigned char md4_hash[21];
-    int userlen=strlen(username);
-    int phase3len=s_min(88+userlen, sizeof phase3);
+    unsigned int userlen=strlen(username);
+    unsigned int phase3len=s_min(88+userlen, sizeof phase3);
 
     /* setup phase3 structure */
     memset(phase3, 0, sizeof phase3);
@@ -531,13 +559,15 @@ static char *ntlm3(char *username, char *password, char *phase2) {
 
     /* decode challenge and calculate response */
     decoded=base64(0, phase2, strlen(phase2)); /* decode */
+    if(!decoded)
+        return NULL;
     crypt_DES((unsigned char *)phase3+64,
         (unsigned char *)decoded+24, md4_hash);
     crypt_DES((unsigned char *)phase3+72,
         (unsigned char *)decoded+24, md4_hash+7);
     crypt_DES((unsigned char *)phase3+80,
         (unsigned char *)decoded+24, md4_hash+14);
-    free(decoded);
+    str_free(decoded);
 
     strncpy(phase3+88, username, sizeof phase3-88);
 
@@ -565,13 +595,22 @@ static void crypt_DES(DES_cblock dst, const_DES_cblock src, DES_cblock hash) {
         (DES_cblock *)dst, &sched, DES_ENCRYPT);
 }
 
+#endif
+
 static char *base64(int encode, char *in, int len) {
     BIO *bio, *b64;
     char *out;
+    int n;
 
     b64=BIO_new(BIO_f_base64());
+    if(!b64)
+        return NULL;
     BIO_set_flags(b64, BIO_FLAGS_BASE64_NO_NL);
     bio=BIO_new(BIO_s_mem());
+    if(!bio) {
+        str_free(b64);
+        return NULL;
+    }
     if(encode)
         bio=BIO_push(b64, bio);
     BIO_write(bio, in, len);
@@ -582,15 +621,23 @@ static char *base64(int encode, char *in, int len) {
     } else {
         bio=BIO_push(b64, bio);
     }
-    len=BIO_pending(bio);
-    out=calloc(len+1, 1);
+    n=BIO_pending(bio);
+    /* 32 bytes as a safety precaution for passing decoded data to crypt_DES */
+    /* n+1 to get null-terminated string on encode */
+    out=str_alloc(n<32?32:n+1);
     if(!out) {
-        s_log(LOG_RAW, "Fatal memory allocation error");
-        die(2);
+        s_log(LOG_ERR, "Memory allocation error");
+        BIO_free_all(bio);
+        return NULL;
+    }
+    n=BIO_read(bio, out, n);
+    if(n<0) {
+        BIO_free_all(bio);
+        str_free(out);
+        return NULL;
     }
-    BIO_read(bio, out, len);
     BIO_free_all(bio);
     return out;
 }
 
-/* End of protocol.c */
+/* end of protocol.c */
