@@ -1,6 +1,6 @@
 /*
  *   stunnel       Universal SSL tunnel
- *   Copyright (C) 1998-2009 Michal Trojnara <Michal.Trojnara@mirt.net>
+ *   Copyright (C) 1998-2011 Michal Trojnara <Michal.Trojnara@mirt.net>
  *
  *   This program is free software; you can redistribute it and/or modify it
  *   under the terms of the GNU General Public License as published by the
@@ -40,10 +40,12 @@
 
 #include "common.h"
 
-/**************************************** Network data structure */
+/**************************************** data structures */
 
 #define MAX_HOSTS 16
 
+typedef enum {LOG_MODE_NONE, LOG_MODE_ERROR, LOG_MODE_FULL} LOG_MODE;
+
 typedef union sockaddr_union {
     struct sockaddr sa;
     struct sockaddr_in in;
@@ -58,71 +60,10 @@ typedef struct sockaddr_list {      /* list of addresses */
     u16 num;                        /* how many addresses are used */
 } SOCKADDR_LIST;
 
-#ifdef __INNOTEK_LIBC__
-#define socklen_t    __socklen_t
-#define strcasecmp   stricmp
-#define strncasecmp  strnicmp
-#define NI_NUMERICHOST 1
-#define NI_NUMERICSERV 2
-#endif
-
-
-/**************************************** Prototypes for stunnel.c */
-
-extern volatile int num_clients;
-
-void main_initialize(char *, char *);
-void main_execute(void);
-#if !defined (USE_WIN32) && !defined (__vms) && !defined(USE_OS2)
-void drop_privileges(void);
-#endif
-void stunnel_info(int);
-void die(int);
-
-/**************************************** Prototypes for log.c */
-
-void log_open(void);
-void log_close(void);
-void log_flush(void);
-void s_log(int, const char *, ...)
-#ifdef __GNUC__
-    __attribute__ ((format (printf, 2, 3)));
-#else
-    ;
-#endif
-void ioerror(const char *);
-void sockerror(const char *);
-void log_error(int, int, const char *);
-char *my_strerror(int);
-
-/**************************************** Prototypes for pty.c */
-/* Based on Public Domain code by Tatu Ylonen <ylo@cs.hut.fi>  */
-
-int pty_allocate(int *, int *, char *, int);
-#if 0
-void pty_release(char *);
-void pty_make_controlling_tty(int *, char *);
-#endif
-
-/**************************************** Prototypes for ssl.c */
-
 typedef enum {
     COMP_NONE, COMP_ZLIB, COMP_RLE
 } COMP_TYPE;
 
-extern int cli_index, opt_index;;
-
-void ssl_init(void);
-void ssl_configure(void);
-#ifdef HAVE_OSSL_ENGINE_H
-void open_engine(const char *);
-void ctrl_engine(const char *, const char *);
-void close_engine(void);
-ENGINE *get_engine(int);
-#endif
-
-/**************************************** Prototypes for options.c */
-
 typedef struct {
         /* some data for SSL initialization in ssl.c */
     COMP_TYPE compression;                               /* compression type */
@@ -140,11 +81,6 @@ typedef struct {
     int uid, gid;
 #endif
 
-        /* Win32 specific data for gui.c */
-#if defined(USE_WIN32) && !defined(_WIN32_WCE)
-    char *win32_service;
-#endif
-
         /* logging-support data for log.c */
     int debug_level;                              /* debug level for logging */
 #ifndef USE_WIN32
@@ -167,82 +103,105 @@ typedef struct {
     } option;
 } GLOBAL_OPTIONS;
 
-extern GLOBAL_OPTIONS options;
+extern GLOBAL_OPTIONS global_options;
 
-typedef struct local_options {
+typedef struct servername_list_struct SERVERNAME_LIST;/* forward declaration */
+
+typedef struct service_options_struct {
+    struct service_options_struct *next;   /* next node in the services list */
     SSL_CTX *ctx;                                            /*  SSL context */
-    X509_STORE *revocation_store;             /* cert store for CRL checking */
-#ifdef HAVE_OSSL_ENGINE_H
-    ENGINE *engine;                        /* engine to read the private key */
-#endif
-    struct local_options *next;            /* next node in the services list */
     char *servname;        /* service name for logging & permission checking */
-    SSL_SESSION *session;                           /* jecently used session */
-    char local_address[IPLEN];             /* dotted-decimal address to bind */
+
+        /* service-specific data for sthreads.c */
 #ifndef USE_FORK
     int stack_size;                            /* stack size for this thread */
 #endif
 
-        /* service-specific data for ctx.c */
+        /* service-specific data for verify.c */
     char *ca_dir;                              /* directory for hashed certs */
     char *ca_file;                       /* file containing bunches of certs */
     char *crl_dir;                              /* directory for hashed CRLs */
     char *crl_file;                       /* file containing bunches of CRLs */
+    int verify_level;
+    X509_STORE *revocation_store;             /* cert store for CRL checking */
+    SOCKADDR_LIST ocsp_addr;
+    char *ocsp_path;
+    unsigned long ocsp_flags;
+
+        /* service-specific data for ctx.c */
     char *cipher_list;
     char *cert;                                             /* cert filename */
     char *key;                               /* pem (priv key/cert) filename */
     long session_timeout;
-    int verify_level;
-    int verify_use_only_my;
     long ssl_options;
-#if SSLEAY_VERSION_NUMBER >= 0x00907000L
-    SOCKADDR_LIST ocsp_addr;
-    char *ocsp_path;
-    unsigned long ocsp_flags;
-#endif /* OpenSSL-0.9.7 */
     SSL_METHOD *client_method, *server_method;
     SOCKADDR_LIST sessiond_addr;
+    SERVERNAME_LIST *servername_list_head, *servername_list_tail;
+    int curve;
+#ifdef HAVE_OSSL_ENGINE_H
+    ENGINE *engine;                        /* engine to read the private key */
+#endif
 
         /* service-specific data for client.c */
     int fd;        /* file descriptor accepting connections for this service */
-    char *execname, **execargs; /* program name and arguments for local mode */
+    SSL_SESSION *session;                           /* recently used session */
+    char *execname;                           /* program name for local mode */
+#ifdef USE_WIN32
+    char *execargs;                      /* program arguments for local mode */
+#else
+    char **execargs;                     /* program arguments for local mode */
+#endif
     SOCKADDR_LIST local_addr, remote_addr, source_addr;
     char *username;
     char *remote_address;
-    int timeout_busy; /* maximum waiting for data time */
-    int timeout_close; /* maximum close_notify time */
-    int timeout_connect; /* maximum connect() time */
-    int timeout_idle; /* maximum idle connection time */
-    enum {FAILOVER_RR, FAILOVER_PRIO} failover; /* failover strategy */
+    char *host_name;
+    int timeout_busy;                       /* maximum waiting for data time */
+    int timeout_close;                          /* maximum close_notify time */
+    int timeout_connect;                           /* maximum connect() time */
+    int timeout_idle;                        /* maximum idle connection time */
+    enum {FAILOVER_RR, FAILOVER_PRIO} failover;         /* failover strategy */
 
-        /* protocol name for protocol.c */
+        /* service-specific data for protocol.c */
     char *protocol;
     char *protocol_host;
     char *protocol_username;
     char *protocol_password;
     char *protocol_authentication;
 
+        /* service-specific data for gui.c */
+#ifdef USE_WIN32
+    int section_number;
+#endif
+
         /* on/off switches */
     struct {
-        unsigned int cert:1;
+        unsigned int accept:1;          /* endpoint: accept */
         unsigned int client:1;
         unsigned int delayed_lookup:1;
-        unsigned int accept:1;
-        unsigned int remote:1;
-        unsigned int retry:1; /* loop remote+program */
+#ifdef USE_LIBWRAP
+        unsigned int libwrap:1;
+#endif
+        unsigned int remote:1;          /* endpoint: connect */
+        unsigned int retry:1;           /* loop remote+program */
         unsigned int sessiond:1;
+        unsigned int program:1;         /* endpoint: exec */
+        unsigned int sni:1;             /* endpoint: sni */
 #ifndef USE_WIN32
-        unsigned int program:1;
         unsigned int pty:1;
-        unsigned int transparent:1;
+        unsigned int transparent_src:1;
+        unsigned int transparent_dst:1; /* endpoint: transparent destination */
 #endif
-#if SSLEAY_VERSION_NUMBER >= 0x00907000L
         unsigned int ocsp:1;
-#endif
     } option;
-} LOCAL_OPTIONS;
+} SERVICE_OPTIONS;
 
-extern LOCAL_OPTIONS local_options;
+extern SERVICE_OPTIONS service_options;
+
+struct servername_list_struct {
+    char *servername;
+    SERVICE_OPTIONS *opt;
+    struct servername_list_struct *next;
+};
 
 typedef enum {
     TYPE_NONE, TYPE_FLAG, TYPE_INT, TYPE_LINGER, TYPE_TIMEVAL, TYPE_STRING
@@ -264,18 +223,11 @@ typedef struct {
     OPT_UNION *opt_val[3];
 } SOCK_OPT;
 
-void parse_config(char *, char *);
+typedef enum {
+    CONF_RELOAD, CONF_FILE, CONF_FD
+} CONF_TYPE;
 
-/**************************************** Prototypes for ctx.c */
-
-void context_init(LOCAL_OPTIONS *);
-void sslerror(char *);
-
-/**************************************** Prototypes for verify.c */
-
-void verify_init(LOCAL_OPTIONS *);
-
-/**************************************** Prototypes for network.c */
+        /* s_poll_set definition for network.c */
 
 #ifdef USE_POLL
 #define MAX_FD 256
@@ -291,66 +243,155 @@ typedef struct {
 #endif
 } s_poll_set;
 
+typedef struct disk_file {
+#ifdef USE_WIN32
+    HANDLE fh;
+#else
+    int fd;
+#endif
+    /* the inteface is prepared to easily implement buffering if needed */
+} DISK_FILE;
+
+/**************************************** prototypes for stunnel.c */
+
+extern volatile int num_clients;
+
+void main_initialize(char *, char *);
+void daemon_loop(void);
+void unbind_ports(void);
+int bind_ports(void);
+#if !defined (USE_WIN32) && !defined (__vms) && !defined(USE_OS2)
+void drop_privileges(void);
+#endif
+int s_socket(int, int, int, int, char *);
+int s_pipe(int [2], int, char *);
+int s_socketpair(int, int, int, int [2], int, char *);
+int s_accept(int, struct sockaddr *, socklen_t *, int, char *);
+void stunnel_info(int);
+void die(int);
+void set_nonblock(int, unsigned long);
+
+/**************************************** prototypes for log.c */
+
+#if !defined(USE_WIN32) && !defined(__vms)
+void syslog_open(void);
+void syslog_close(void);
+#endif
+void log_open(void);
+void log_close(void);
+void log_flush(LOG_MODE);
+void s_log(int, const char *, ...)
+#ifdef __GNUC__
+    __attribute__ ((format (printf, 2, 3)));
+#else
+    ;
+#endif
+void ioerror(const char *);
+void sockerror(const char *);
+void log_error(int, int, const char *);
+char *s_strerror(int);
+
+/**************************************** prototypes for pty.c */
+
+int pty_allocate(int *, int *, char *);
+
+/**************************************** prototypes for ssl.c */
+
+extern int cli_index, opt_index;
+
+void ssl_init(void);
+int ssl_configure(void);
+#ifdef HAVE_OSSL_ENGINE_H
+char *open_engine(const char *);
+char *ctrl_engine(const char *, const char *);
+void close_engine(void);
+ENGINE *get_engine(int);
+#endif
+
+/**************************************** prototypes for options.c */
+
+void parse_commandline(char *, char *);
+int parse_conf(char *, CONF_TYPE);
+
+/**************************************** prototypes for ctx.c */
+
+int context_init(SERVICE_OPTIONS *);
+void sslerror(char *);
+
+/**************************************** prototypes for verify.c */
+
+int verify_init(SERVICE_OPTIONS *);
+
+/**************************************** prototypes for network.c */
+
 void s_poll_init(s_poll_set *);
 void s_poll_add(s_poll_set *, int, int, int);
 int s_poll_canread(s_poll_set *, int);
 int s_poll_canwrite(s_poll_set *, int);
+int s_poll_error(s_poll_set *, int);
 int s_poll_wait(s_poll_set *, int, int);
 
-#ifndef USE_WIN32
+#ifdef USE_WIN32
+#define SIGNAL_RELOAD_CONFIG    1
+#define SIGNAL_REOPEN_LOG       2
+#define SIGNAL_TERMINATE        3
+#else
+#define SIGNAL_RELOAD_CONFIG    SIGHUP
+#define SIGNAL_REOPEN_LOG       SIGUSR1
+#define SIGNAL_TERMINATE        SIGTERM
+#endif
+void signal_handler(int);
 int signal_pipe_init(void);
+void signal_post(int);
+#if !defined(USE_WIN32) && !defined(USE_OS2)
 void child_status(void);  /* dead libwrap or 'exec' process detected */
 #endif
-int set_socket_options(int, int);
-int alloc_fd(int);
-void setnonblock(int, unsigned long);
 
-/**************************************** Prototypes for client.c */
+int set_socket_options(int, int);
+int get_socket_error(const int);
+int make_sockets(int [2]);
+
+/**************************************** prototypes for client.c */
 
 typedef struct {
-    int fd; /* File descriptor */
-    int rd; /* Open for read */
-    int wr; /* Open for write */
-    int is_socket; /* File descriptor is a socket */
+    int fd; /* file descriptor */
+    int is_socket; /* file descriptor is a socket */
 } FD;
 
 typedef struct {
-    LOCAL_OPTIONS *opt;
-    char accepted_address[IPLEN]; /* text */
-    SOCKADDR_LIST peer_addr; /* Peer address */
-    FD local_rfd, local_wfd; /* Read and write local descriptors */
-    FD remote_fd; /* Remote file descriptor */
-    SSL *ssl; /* SSL Connection */
+    SSL *ssl; /* SSL connnection */
+    SERVICE_OPTIONS *opt;
+    jmp_buf err; /* exception handler */
+
+    char accepted_address[IPLEN]; /* IP address as text for logging */
+    SOCKADDR_LIST peer_addr; /* peer address */
+    FD local_rfd, local_wfd; /* read and write local descriptors */
+    FD remote_fd; /* remote file descriptor */
     SOCKADDR_LIST bind_addr;
         /* IP for explicit local bind or transparent proxy */
-    unsigned long pid; /* PID of local process */
-    int fd; /* Temporary file descriptor */
-    jmp_buf err;
+    unsigned long pid; /* PID of the local process */
+    int fd; /* temporary file descriptor */
 
-    char sock_buff[BUFFSIZE]; /* Socket read buffer */
+    /* data for transfer() function */
+    char sock_buff[BUFFSIZE]; /* socket read buffer */
     char ssl_buff[BUFFSIZE]; /* SSL read buffer */
-    int sock_ptr, ssl_ptr; /* Index of first unused byte in buffer */
-    FD *sock_rfd, *sock_wfd; /* Read and write socket descriptors */
-    FD *ssl_rfd, *ssl_wfd; /* Read and write SSL descriptors */
-    int sock_bytes, ssl_bytes; /* Bytes written to socket and ssl */
-    s_poll_set fds; /* File descriptors */
+    int sock_ptr, ssl_ptr; /* index of first unused byte in buffer */
+    FD *sock_rfd, *sock_wfd; /* read and write socket descriptors */
+    FD *ssl_rfd, *ssl_wfd; /* read and write SSL descriptors */
+    int sock_bytes, ssl_bytes; /* bytes written to socket and SSL */
+    s_poll_set fds; /* file descriptors */
 } CLI;
 
-extern int max_clients;
-#ifndef USE_WIN32
-extern int max_fds;
-#endif
-
-CLI *alloc_client_session(LOCAL_OPTIONS *, int, int);
+CLI *alloc_client_session(SERVICE_OPTIONS *, int, int);
 void *client(void *);
 
-/**************************************** Prototypes for network.c */
+/**************************************** prototypes for network.c */
 
 int connect_blocking(CLI *, SOCKADDR_UNION *, socklen_t);
 void write_blocking(CLI *, int fd, void *, int);
 void read_blocking(CLI *, int fd, void *, int);
 void fdputline(CLI *, int, const char *);
-void fdgetline(CLI *, int, char *);
+char *fdgetline(CLI *, int);
 /* descriptor versions of fprintf/fscanf */
 int fdprintf(CLI *, int, const char *, ...)
 #ifdef __GNUC__
@@ -358,28 +399,31 @@ int fdprintf(CLI *, int, const char *, ...)
 #else
        ;
 #endif
-int fdscanf(CLI *, int, const char *, char *)
-#ifdef __GNUC__
-       __attribute__ ((format (scanf, 3, 0)));
-#else
-       ;
-#endif
 
-/**************************************** Prototype for protocol.c */
+/**************************************** prototype for protocol.c */
 
 void negotiate(CLI *c);
 
-/**************************************** Prototypes for resolver.c */
+/**************************************** prototypes for resolver.c */
 
 int name2addrlist(SOCKADDR_LIST *, char *, char *);
 int hostport2addrlist(SOCKADDR_LIST *, char *, char *);
 char *s_ntop(char *, SOCKADDR_UNION *);
 
-/**************************************** Prototypes for sthreads.c */
+/**************************************** prototypes for sthreads.c */
 
 typedef enum {
-    CRIT_KEYGEN, CRIT_INET, CRIT_CLIENTS, CRIT_WIN_LOG, CRIT_SESSION,
-    CRIT_LIBWRAP, CRIT_SSL, CRIT_SECTIONS
+    CRIT_CLIENTS, CRIT_SESSION, /* client.c */
+#if OPENSSL_VERSION_NUMBER<0x1000002f
+    CRIT_SSL,                   /* client.c */
+#endif /* OpenSSL version < 1.0.0b */
+    CRIT_INET,                  /* resolver.c */
+#ifdef USE_WIN32
+    CRIT_WIN_LOG,               /* gui.c */
+#else
+    CRIT_LIBWRAP,               /* libwrap.c */
+#endif
+    CRIT_SECTIONS               /* number of critical sections */
 } SECTION_CODE;
 
 void enter_critical_section(SECTION_CODE);
@@ -392,33 +436,36 @@ int create_client(int, int, CLI *, void *(*)(void *));
 typedef struct CONTEXT_STRUCTURE {
     char *stack; /* CPU stack for this thread */
     unsigned long id;
-    ucontext_t ctx;
+    ucontext_t context;
     s_poll_set *fds;
     int ready; /* number of ready file descriptors */
     time_t finish; /* when to finish poll() for this context */
     struct CONTEXT_STRUCTURE *next; /* next context on a list */
+    void *tls; /* thread local storage for str.c */
 } CONTEXT;
 extern CONTEXT *ready_head, *ready_tail;
 extern CONTEXT *waiting_head, *waiting_tail;
 #endif
 #ifdef _WIN32_WCE
-int _beginthread(void (*)(void *), int, void *);
+long _beginthread(void (*)(void *), int, void *);
 void _endthread(void);
 #endif
 #ifdef DEBUG_STACK_SIZE
 void stack_info(int);
 #endif
 
-/**************************************** Prototypes for gui.c */
+/**************************************** prototypes for gui.c */
 
 typedef struct {
-    LOCAL_OPTIONS *section;
+    SERVICE_OPTIONS *section;
     char pass[PEM_BUFSIZE];
 } UI_DATA;
 
 #ifdef USE_WIN32
 void win_log(char *);
-void exit_win32(int);
+void win_exit(int);
+void win_newconfig(int);
+void win_newcert(SSL *, SERVICE_OPTIONS *);
 int passwd_cb(char *, int, int, void *);
 #ifdef HAVE_OSSL_ENGINE_H
 int pin_cb(UI *, UI_STRING *);
@@ -436,16 +483,7 @@ extern GETNAMEINFO s_getnameinfo;
 #endif /* ! _WIN32_WCE */
 #endif /* USE_WIN32 */
 
-/**************************************** Prototypes for file.c */
-
-typedef struct disk_file {
-#ifdef USE_WIN32
-    HANDLE fh;
-#else
-    int fd;
-#endif
-    /* the inteface is prepared to easily implement buffering if needed */
-} DISK_FILE;
+/**************************************** prototypes for file.c */
 
 #ifndef USE_WIN32
 DISK_FILE *file_fdopen(int);
@@ -460,11 +498,29 @@ LPTSTR str2tstr(const LPSTR);
 LPSTR tstr2str(const LPTSTR);
 #endif
 
-/**************************************** Prototypes for libwrap.c */
+/**************************************** prototypes for libwrap.c */
 
 void libwrap_init(int);
-void auth_libwrap(CLI *);
+void libwrap_auth(CLI *);
+
+/**************************************** prototypes for str.c */
+
+void str_init();
+void str_cleanup();
+void str_stats();
+void *str_alloc(size_t);
+void *str_realloc(void *, size_t);
+void str_detach(void *);
+void str_free(void *);
+char *str_dup(const char *);
+char *str_vprintf(const char *, va_list);
+char *str_printf(const char *, ...)
+#ifdef __GNUC__
+    __attribute__ ((format (printf, 1, 2)));
+#else
+    ;
+#endif
 
 #endif /* defined PROTOTYPES_H */
 
-/* End of prototypes.h */
+/* end of prototypes.h */
