@@ -1,6 +1,6 @@
 /*
  *   stunnel       Universal SSL tunnel
- *   Copyright (C) 1998-2009 Michal Trojnara <Michal.Trojnara@mirt.net>
+ *   Copyright (C) 1998-2011 Michal Trojnara <Michal.Trojnara@mirt.net>
  *
  *   This program is free software; you can redistribute it and/or modify it
  *   under the terms of the GNU General Public License as published by the
@@ -47,10 +47,12 @@
 /* no need for critical sections */
 
 void enter_critical_section(SECTION_CODE i) {
+    (void)i; /* skip warning about unused parameter */
     /* empty */
 }
 
 void leave_critical_section(SECTION_CODE i) {
+    (void)i; /* skip warning about unused parameter */
     /* empty */
 }
 
@@ -76,7 +78,6 @@ void leave_critical_section(SECTION_CODE i) {
 /* first context on the ready list is the active context */
 CONTEXT *ready_head=NULL, *ready_tail=NULL;         /* ready to execute */
 CONTEXT *waiting_head=NULL, *waiting_tail=NULL;     /* waiting on poll() */
-int next_id=1;
 
 unsigned long stunnel_process_id(void) {
     return (unsigned long)getpid();
@@ -86,66 +87,91 @@ unsigned long stunnel_thread_id(void) {
     return ready_head ? ready_head->id : 0;
 }
 
-static CONTEXT *new_context(int stack_size) {
-    CONTEXT *ctx;
+static CONTEXT *new_context(void) {
+    static int next_id=1;
+    CONTEXT *context;
 
     /* allocate and fill the CONTEXT structure */
-    ctx=malloc(sizeof(CONTEXT));
-    if(!ctx) {
+    context=str_alloc(sizeof(CONTEXT));
+    if(!context) {
         s_log(LOG_ERR, "Unable to allocate CONTEXT structure");
         return NULL;
     }
-    ctx->stack=malloc(stack_size);
-    if(!ctx->stack) {
-        s_log(LOG_ERR, "Unable to allocate CONTEXT stack");
-        return NULL;
-    }
-    ctx->id=next_id++;
-    ctx->fds=NULL;
-    ctx->ready=0;
-    /* some manuals claim that initialization of ctx structure is required */
-    if(getcontext(&ctx->ctx)<0) {
-        free(ctx->stack);
-        free(ctx);
-        ioerror("getcontext");
-        return NULL;
-    }
-    ctx->ctx.uc_link=NULL; /* it should never happen */
-#if defined(__sgi) || ARGC==2 /* obsolete ss_sp semantics */
-    ctx->ctx.uc_stack.ss_sp=ctx->stack+stack_size-8;
-#else
-    ctx->ctx.uc_stack.ss_sp=ctx->stack;
-#endif
-    ctx->ctx.uc_stack.ss_size=stack_size;
-    ctx->ctx.uc_stack.ss_flags=0;
+    str_detach(context);
+    context->id=next_id++;
+    context->fds=NULL;
+    context->ready=0;
 
-    /* attach to the tail of the ready queue */
-    ctx->next=NULL;
+    /* append to the tail of the ready queue */
+    context->next=NULL;
     if(ready_tail)
-        ready_tail->next=ctx;
-    ready_tail=ctx;
+        ready_tail->next=context;
+    ready_tail=context;
     if(!ready_head)
-        ready_head=ctx;
-    return ctx;
+        ready_head=context;
+
+    return context;
 }
 
 void sthreads_init(void) {
     /* create the first (listening) context and put it in the running queue */
-    if(!new_context(DEFAULT_STACK_SIZE)) {
-        s_log(LOG_RAW, "Unable create the listening context");
+    if(!new_context()) {
+        s_log(LOG_ERR, "Unable create the listening context");
         die(1);
     }
+    /* no need to initialize ucontext_t structure here
+       it will be initialied with swapcontext() call */
 }
 
 int create_client(int ls, int s, CLI *arg, void *(*cli)(void *)) {
-    CONTEXT *ctx;
+    CONTEXT *context;
 
+    (void)ls; /* this parameter is only used with USE_FORK */
+    
     s_log(LOG_DEBUG, "Creating a new context");
-    ctx=new_context(arg->opt->stack_size);
-    if(!ctx)
+    context=new_context();
+    if(!context) {
+        if(arg)
+            str_free(arg);
+        if(s>=0)
+            closesocket(s);
         return -1;
-    s_log(LOG_DEBUG, "Context %ld created", ctx->id);
-    makecontext(&ctx->ctx, (void(*)(void))cli, ARGC, arg);
+    }
+
+    /* initialize context_t structure */
+    if(getcontext(&context->context)<0) {
+        str_free(context);
+        if(arg)
+            str_free(arg);
+        if(s>=0)
+            closesocket(s);
+        ioerror("getcontext");
+        return -1;
+    }
+    context->context.uc_link=NULL; /* stunnel does not use uc_link */
+
+    /* create stack */
+    context->stack=str_alloc(arg->opt->stack_size);
+    if(!context->stack) {
+        str_free(context);
+        if(arg)
+            str_free(arg);
+        if(s>=0)
+            closesocket(s);
+        s_log(LOG_ERR, "Unable to allocate stack");
+        return -1;
+    }
+    str_detach(context->stack);
+#if defined(__sgi) || ARGC==2 /* obsolete ss_sp semantics */
+    context->context.uc_stack.ss_sp=context->stack+arg->opt->stack_size-8;
+#else
+    context->context.uc_stack.ss_sp=context->stack;
+#endif
+    context->context.uc_stack.ss_size=arg->opt->stack_size;
+    context->context.uc_stack.ss_flags=0;
+
+    s_log(LOG_DEBUG, "Context %ld created", context->id);
+    makecontext(&context->context, (void(*)(void))cli, ARGC, arg);
     return 0;
 }
 
@@ -173,7 +199,7 @@ int create_client(int ls, int s, CLI *arg, void *(*cli)(void *)) {
     switch(fork()) {
     case -1:    /* error */
         if(arg)
-            free(arg);
+            str_free(arg);
         if(s>=0)
             closesocket(s);
         return -1;
@@ -185,7 +211,7 @@ int create_client(int ls, int s, CLI *arg, void *(*cli)(void *)) {
         _exit(0);
     default:    /* parent */
         if(arg)
-            free(arg);
+            str_free(arg);
         if(s>=0)
             closesocket(s);
     }
@@ -209,9 +235,11 @@ void leave_critical_section(SECTION_CODE i) {
 
 static void locking_callback(int mode, int type,
 #ifdef HAVE_OPENSSL
-    const /* Callback definition has been changed in openssl 0.9.3 */
+    const /* callback definition has been changed in openssl 0.9.3 */
 #endif
     char *file, int line) {
+    (void)file; /* skip warning about unused parameter */
+    (void)line; /* skip warning about unused parameter */
     if(mode&CRYPTO_LOCK)
         pthread_mutex_lock(lock_cs+type);
     else
@@ -226,15 +254,20 @@ static struct CRYPTO_dynlock_value *dyn_create_function(const char *file,
         int line) {
     struct CRYPTO_dynlock_value *value;
 
-    value=malloc(sizeof(struct CRYPTO_dynlock_value));
+    (void)file; /* skip warning about unused parameter */
+    (void)line; /* skip warning about unused parameter */
+    value=str_alloc(sizeof(struct CRYPTO_dynlock_value));
     if(!value)
         return NULL;
+    str_detach(value);
     pthread_mutex_init(&value->mutex, NULL);
     return value;
 }
 
 static void dyn_lock_function(int mode, struct CRYPTO_dynlock_value *value,
         const char *file, int line) {
+    (void)file; /* skip warning about unused parameter */
+    (void)line; /* skip warning about unused parameter */
     if(mode&CRYPTO_LOCK)
         pthread_mutex_lock(&value->mutex);
     else
@@ -243,8 +276,10 @@ static void dyn_lock_function(int mode, struct CRYPTO_dynlock_value *value,
 
 static void dyn_destroy_function(struct CRYPTO_dynlock_value *value,
         const char *file, int line) {
+    (void)file; /* skip warning about unused parameter */
+    (void)line; /* skip warning about unused parameter */
     pthread_mutex_destroy(&value->mutex);
-    free(value);
+    str_free(value);
 }
 
 unsigned long stunnel_process_id(void) {
@@ -258,54 +293,59 @@ unsigned long stunnel_thread_id(void) {
 void sthreads_init(void) {
     int i;
 
-    /* Initialize stunnel critical sections */
+    /* initialize stunnel critical sections */
     for(i=0; i<CRIT_SECTIONS; i++)
         pthread_mutex_init(stunnel_cs+i, NULL);
 
-    /* Initialize OpenSSL locking callback */
+    /* initialize OpenSSL locking callback */
     for(i=0; i<CRYPTO_NUM_LOCKS; i++)
         pthread_mutex_init(lock_cs+i, NULL);
     CRYPTO_set_id_callback(stunnel_thread_id);
     CRYPTO_set_locking_callback(locking_callback);
 
-    /* Initialize OpenSSL dynamic locks callbacks */
+    /* initialize OpenSSL dynamic locks callbacks */
     CRYPTO_set_dynlock_create_callback(dyn_create_function);
     CRYPTO_set_dynlock_lock_callback(dyn_lock_function);
     CRYPTO_set_dynlock_destroy_callback(dyn_destroy_function);
 }
 
 int create_client(int ls, int s, CLI *arg, void *(*cli)(void *)) {
-    pthread_attr_t pth_attr;
     pthread_t thread;
-#ifdef HAVE_PTHREAD_SIGMASK
-    sigset_t newmask, oldmask;
+    pthread_attr_t pth_attr;
+    int error;
+#if defined(HAVE_PTHREAD_SIGMASK) && !defined(__APPLE__)
+    /* Disabled on OS X due to strange problems on Mac OS X 10.5
+       it seems to restore signal mask somewhere (I couldn't find where)
+       effectively blocking signals after first accepted connection */
+    sigset_t new_set, old_set;
+#endif /* HAVE_PTHREAD_SIGMASK && !__APPLE__*/
 
-    /* Initialize attributes for creating new threads */
+    (void)ls; /* this parameter is only used with USE_FORK */
+
+#if defined(HAVE_PTHREAD_SIGMASK) && !defined(__APPLE__)
+    /* the idea is that only the main thread handles all the signals with
+     * posix threads;  signals are blocked for any other thread */
+    sigfillset(&new_set);
+    pthread_sigmask(SIG_SETMASK, &new_set, &old_set); /* block signals */
+#endif /* HAVE_PTHREAD_SIGMASK && !__APPLE__*/
     pthread_attr_init(&pth_attr);
     pthread_attr_setdetachstate(&pth_attr, PTHREAD_CREATE_DETACHED);
     pthread_attr_setstacksize(&pth_attr, arg->opt->stack_size);
+    error=pthread_create(&thread, &pth_attr, cli, arg);
+    pthread_attr_destroy(&pth_attr);
+#if defined(HAVE_PTHREAD_SIGMASK) && !defined(__APPLE__)
+    pthread_sigmask(SIG_SETMASK, &old_set, NULL); /* unblock signals */
+#endif /* HAVE_PTHREAD_SIGMASK && !__APPLE__*/
 
-    /* The idea is that only the main thread handles all the signals with
-     * posix threads.  Signals are blocked for any other thread. */
-    sigemptyset(&newmask);
-    sigaddset(&newmask, SIGCHLD);
-    sigaddset(&newmask, SIGTERM);
-    sigaddset(&newmask, SIGQUIT);
-    sigaddset(&newmask, SIGINT);
-    sigaddset(&newmask, SIGHUP);
-    pthread_sigmask(SIG_BLOCK, &newmask, &oldmask); /* block signals */
-#endif /* HAVE_PTHREAD_SIGMASK */
-    if(pthread_create(&thread, &pth_attr, cli, arg)) {
-#ifdef HAVE_PTHREAD_SIGMASK
-        pthread_sigmask(SIG_SETMASK, &oldmask, NULL); /* restore the mask */
-#endif /* HAVE_PTHREAD_SIGMASK */
+    if(error) {
+        errno=error;
+        ioerror("pthread_create");
+        if(arg)
+            str_free(arg);
         if(s>=0)
             closesocket(s);
         return -1;
     }
-#ifdef HAVE_PTHREAD_SIGMASK
-    pthread_sigmask(SIG_SETMASK, &oldmask, NULL); /* restore the mask */
-#endif /* HAVE_PTHREAD_SIGMASK */
     return 0;
 }
 
@@ -326,9 +366,11 @@ void leave_critical_section(SECTION_CODE i) {
 
 static void locking_callback(int mode, int type,
 #ifdef HAVE_OPENSSL
-    const /* Callback definition has been changed in openssl 0.9.3 */
+    const /* callback definition has been changed in openssl 0.9.3 */
 #endif
     char *file, int line) {
+    (void)file; /* skip warning about unused parameter */
+    (void)line; /* skip warning about unused parameter */
     if(mode&CRYPTO_LOCK)
         EnterCriticalSection(lock_cs+type);
     else
@@ -343,15 +385,20 @@ static struct CRYPTO_dynlock_value *dyn_create_function(const char *file,
         int line) {
     struct CRYPTO_dynlock_value *value;
 
-    value=malloc(sizeof(struct CRYPTO_dynlock_value));
+    (void)file; /* skip warning about unused parameter */
+    (void)line; /* skip warning about unused parameter */
+    value=str_alloc(sizeof(struct CRYPTO_dynlock_value));
     if(!value)
         return NULL;
+    str_detach(value);
     InitializeCriticalSection(&value->mutex);
     return value;
 }
 
 static void dyn_lock_function(int mode, struct CRYPTO_dynlock_value *value,
         const char *file, int line) {
+    (void)file; /* skip warning about unused parameter */
+    (void)line; /* skip warning about unused parameter */
     if(mode&CRYPTO_LOCK)
         EnterCriticalSection(&value->mutex);
     else
@@ -360,8 +407,10 @@ static void dyn_lock_function(int mode, struct CRYPTO_dynlock_value *value,
 
 static void dyn_destroy_function(struct CRYPTO_dynlock_value *value,
         const char *file, int line) {
+    (void)file; /* skip warning about unused parameter */
+    (void)line; /* skip warning about unused parameter */
     DeleteCriticalSection(&value->mutex);
-    free(value);
+    str_free(value);
 }
 
 unsigned long stunnel_process_id(void) {
@@ -375,25 +424,30 @@ unsigned long stunnel_thread_id(void) {
 void sthreads_init(void) {
     int i;
 
-    /* Initialize stunnel critical sections */
+    /* initialize stunnel critical sections */
     for(i=0; i<CRIT_SECTIONS; i++)
         InitializeCriticalSection(stunnel_cs+i);
 
-    /* Initialize OpenSSL locking callback */
+    /* initialize OpenSSL locking callback */
     for(i=0; i<CRYPTO_NUM_LOCKS; i++)
         InitializeCriticalSection(lock_cs+i);
     CRYPTO_set_locking_callback(locking_callback);
 
-    /* Initialize OpenSSL dynamic locks callbacks */
+    /* initialize OpenSSL dynamic locks callbacks */
     CRYPTO_set_dynlock_create_callback(dyn_create_function);
     CRYPTO_set_dynlock_lock_callback(dyn_lock_function);
     CRYPTO_set_dynlock_destroy_callback(dyn_destroy_function);
 }
 
 int create_client(int ls, int s, CLI *arg, void *(*cli)(void *)) {
+    (void)ls; /* this parameter is only used with USE_FORK */
     s_log(LOG_DEBUG, "Creating a new thread");
-    if(_beginthread((void(*)(void *))cli, arg->opt->stack_size, arg)==-1) {
+    if((long)_beginthread((void(*)(void *))cli, arg->opt->stack_size, arg)==-1) {
         ioerror("_beginthread");
+        if(arg)
+            str_free(arg);
+        if(s>=0)
+            closesocket(s);
         return -1;
     }
     s_log(LOG_DEBUG, "New thread created");
@@ -428,9 +482,14 @@ unsigned long stunnel_thread_id(void) {
 }
 
 int create_client(int ls, int s, CLI *arg, void *(*cli)(void *)) {
+    (void)ls; /* this parameter is only used with USE_FORK */
     s_log(LOG_DEBUG, "Creating a new thread");
-    if(_beginthread((void(*)(void *))cli, NULL, arg->opt->stack_size, arg)==-1) {
+    if((long)_beginthread((void(*)(void *))cli, NULL, arg->opt->stack_size, arg)==-1L) {
         ioerror("_beginthread");
+        if(arg)
+            str_free(arg);
+        if(s>=0)
+            closesocket(s);
         return -1;
     }
     s_log(LOG_DEBUG, "New thread created");
@@ -441,15 +500,16 @@ int create_client(int ls, int s, CLI *arg, void *(*cli)(void *)) {
 
 #ifdef _WIN32_WCE
 
-int _beginthread(void (*start_address)(void *),
+long _beginthread(void (*start_address)(void *),
         int stack_size, void *arglist) {
     DWORD thread_id;
     HANDLE handle;
 
     handle=CreateThread(NULL, stack_size,
-        (LPTHREAD_START_ROUTINE)start_address, arglist, 0, &thread_id);
+        (LPTHREAD_START_ROUTINE)start_address, arglist,
+        STACK_SIZE_PARAM_IS_A_RESERVATION, &thread_id);
     if(!handle)
-        return -1;
+        return -1L;
     CloseHandle(handle);
     return 0;
 }
@@ -505,4 +565,4 @@ void stack_info(int init) { /* 1-initialize, 0-display */
 
 #endif /* DEBUG_STACK_SIZE */
 
-/* End of sthreads.c */
+/* end of sthreads.c */
