@@ -40,7 +40,6 @@
 
 /**************************************** prototypes */
 
-static void daemon_loop(void);
 static void accept_connection(SERVICE_OPTIONS *);
 static void get_limits(void); /* setup global max_clients and max_fds */
 #if !defined(USE_WIN32) && !defined(__vms)
@@ -56,21 +55,24 @@ static int setup_fd(int, int, char *);
 static int max_fds;
 static int max_clients=0;
 
-int volatile num_clients=0; /* current number of clients */
+volatile int num_clients=0; /* current number of clients */
 s_poll_set fds; /* file descriptors of listening sockets */
-#if !defined(USE_WIN32) && !defined(USE_OS2)
 int signal_fd;
-#endif
 
 /**************************************** startup */
 
 #ifndef USE_WIN32
 int main(int argc, char* argv[]) { /* execution begins here 8-) */
-
+    str_init(); /* initialize per-thread string management */
     main_initialize(argc>1 ? argv[1] : NULL, argc>2 ? argv[2] : NULL);
-
-    main_execute();
-
+    if(service_options.next) { /* there are service sections -> daemon mode */
+        num_clients=0;
+        daemon_loop();
+    } else { /* inetd mode */
+        num_clients=1;
+        client(alloc_client_session(&service_options, 0, 1));
+        log_close();
+    }
     return 0; /* success */
 }
 #endif
@@ -78,10 +80,16 @@ int main(int argc, char* argv[]) { /* execution begins here 8-) */
 void main_initialize(char *arg1, char *arg2) {
     ssl_init(); /* initialize SSL library */
     sthreads_init(); /* initialize critical sections & SSL callbacks */
-    parse_commandline(arg1, arg2);
+    get_limits(); /* required by setup_fd() */
 
-    max_fds=FD_SETSIZE; /* start with select() limit */
-    get_limits();
+    signal_fd=signal_pipe_init();
+    s_poll_init(&fds);
+    s_poll_add(&fds, signal_fd, 1, 0);
+    /* the most essential initialization is performed here,
+     * so gui.c can execute a thread with daemon_loop() */
+
+    stunnel_info(LOG_NOTICE);
+    parse_commandline(arg1, arg2);
 #ifdef USE_LIBWRAP
     /* spawn LIBWRAP_CLIENTS processes unless inetd mode is configured
      * execute after parse_commandline() to know service_options.next,
@@ -93,10 +101,7 @@ void main_initialize(char *arg1, char *arg2) {
      * to be able to access /dev/log socket */
     syslog_open();
 #endif /* !defined(USE_WIN32) && !defined(__vms) */
-#if !defined(USE_WIN32) && !defined(USE_OS2)
-    signal_fd=signal_pipe_init();
-#endif
-    if(!bind_ports())
+    if(bind_ports())
         die(1);
 
 #ifdef HAVE_CHROOT
@@ -126,35 +131,23 @@ void main_initialize(char *arg1, char *arg2) {
         create_pid();
     }
 #endif /* standard Unix */
-
-    stunnel_info(LOG_NOTICE);
-}
-
-void main_execute(void) {
-    if(service_options.next) { /* there are service sections -> daemon mode */
-        num_clients=0;
-        while(1)
-            daemon_loop();
-    } else { /* inetd mode */
-        num_clients=1;
-        client(alloc_client_session(&service_options, 0, 1));
-        log_close();
-    }
 }
 
 /**************************************** main loop */
 
-static void daemon_loop(void) {
+void daemon_loop(void) {
     SERVICE_OPTIONS *opt;
 
-    if(s_poll_wait(&fds, -1, -1)>=0) { /* non-critical error */
-        for(opt=service_options.next; opt; opt=opt->next)
-            if(s_poll_canread(&fds, opt->fd))
-                accept_connection(opt);
-    } else {
-        log_error(LOG_INFO, get_last_socket_error(),
-            "daemon_loop: s_poll_wait");
-        sleep(1); /* to avoid log trashing */
+    while(1) {
+        if(s_poll_wait(&fds, -1, -1)>=0) { /* non-critical error */
+            for(opt=service_options.next; opt; opt=opt->next)
+                if(s_poll_canread(&fds, opt->fd))
+                    accept_connection(opt);
+        } else {
+            log_error(LOG_INFO, get_last_socket_error(),
+                "daemon_loop: s_poll_wait");
+            sleep(1); /* to avoid log trashing */
+        }
     }
 }
 
@@ -195,78 +188,94 @@ static void accept_connection(SERVICE_OPTIONS *opt) {
         closesocket(s);
         return;
     }
+    enter_critical_section(CRIT_CLIENTS); /* for multi-cpu machines */
+    /* increment before create_client() to prevent race condition
+     * resulting in logging "Service xxx finished (-1 left)" */
+    ++num_clients;
+    leave_critical_section(CRIT_CLIENTS);
     if(create_client(opt->fd, s, alloc_client_session(opt, s, s), client)) {
         s_log(LOG_ERR, "Connection rejected: create_client failed");
+        enter_critical_section(CRIT_CLIENTS); /* for multi-cpu machines */
+        --num_clients;
+        leave_critical_section(CRIT_CLIENTS);
         closesocket(s);
         return;
     }
-    enter_critical_section(CRIT_CLIENTS); /* for multi-cpu machines */
-    ++num_clients;
-    leave_critical_section(CRIT_CLIENTS);
 }
 
 /**************************************** initialization helpers */
 
-/* close old ports, open new ports, update fds */
-int bind_ports(void) {
+/* clear fds, close old ports */
+void unbind_ports(void) {
     SERVICE_OPTIONS *opt;
-    static SERVICE_OPTIONS *prev_opt=NULL;
-    SOCKADDR_UNION addr;
 
     s_poll_init(&fds);
-#if !defined(USE_WIN32) && !defined(USE_OS2)
     s_poll_add(&fds, signal_fd, 1, 0);
-#endif
-
-    for(opt=prev_opt; opt; opt=opt->next)
-        if(opt->option.accept) {
+    for(opt=service_options.next; opt; opt=opt->next)
+        if(opt->option.accept && opt->fd>=0) {
             closesocket(opt->fd);
             s_log(LOG_DEBUG, "Service %s closed FD=%d",
                 opt->servname, opt->fd);
+            opt->fd=-1;
         }
-    prev_opt=service_options.next;
+}
 
-    for(opt=prev_opt; opt; opt=opt->next) {
+/* open new ports, update fds */
+int bind_ports(void) {
+    SERVICE_OPTIONS *opt;
+    SOCKADDR_UNION addr;
+    char local_address[IPLEN];
+
+    s_poll_init(&fds);
+    s_poll_add(&fds, signal_fd, 1, 0);
+    for(opt=service_options.next; opt; opt=opt->next) {
         if(opt->option.accept) {
             memcpy(&addr, &opt->local_addr.addr[0], sizeof addr);
             opt->fd=s_socket(addr.sa.sa_family, SOCK_STREAM, 0, 1, "accept socket");
             if(opt->fd<0)
-                return 0;
-            if(set_socket_options(opt->fd, 0)<0)
-                return 0;
-            s_ntop(opt->local_address, &addr);
+                return 1;
+            if(set_socket_options(opt->fd, 0)<0) {
+                closesocket(opt->fd);
+                return 1;
+            }
+            s_ntop(local_address, &addr);
             if(bind(opt->fd, &addr.sa, addr_len(addr))) {
                 s_log(LOG_ERR, "Error binding %s to %s",
-                    opt->servname, opt->local_address);
+                    opt->servname, local_address);
                 sockerror("bind");
-                return 0;
+                closesocket(opt->fd);
+                return 1;
             }
             s_log(LOG_DEBUG, "Service %s bound to %s",
-                opt->servname, opt->local_address);
-            if(listen(opt->fd, 5)) {
+                opt->servname, local_address);
+            if(listen(opt->fd, SOMAXCONN)) {
                 sockerror("listen");
-                return 0;
+                closesocket(opt->fd);
+                return 1;
             }
             s_poll_add(&fds, opt->fd, 1, 0);
             s_log(LOG_DEBUG, "Service %s opened FD=%d",
                 opt->servname, opt->fd);
-        } else { /* create exec+connect services */
+        } else if(opt->option.program && opt->option.remote) {
+            /* create exec+connect services */
             enter_critical_section(CRIT_CLIENTS);
             ++num_clients;
             leave_critical_section(CRIT_CLIENTS);
             create_client(-1, -1, alloc_client_session(opt, -1, -1), client);
         }
     }
-    return 1; /* OK */
+    return 0; /* OK */
 }
 
 static void get_limits(void) {
 #if defined(USE_WIN32) || defined(USE_POLL)
     max_fds=0; /* unlimited */
 #elif defined(USE_OS2) && defined(__INNOTEK_LIBC__)
+    max_fds=0; /* unlimited */
     /* OS/2 with the Innotek LIBC does not share the same
        handles between files and socket connections */
 #else /* Unix */
+    max_fds=FD_SETSIZE; /* start with select() limit */
 #if defined(HAVE_SYSCONF)
     int open_max;
 
@@ -289,10 +298,10 @@ static void get_limits(void) {
 
     if(max_fds) {
         max_clients=max_fds>=256 ? max_fds*125/256 : (max_fds-6)/2;
-        s_log(LOG_NOTICE, "Clients allowed=%d", max_clients);
+        s_log(LOG_DEBUG, "Clients allowed=%d", max_clients);
     } else {
         max_clients=0;
-        s_log(LOG_NOTICE, "No limit detected for the number of clients");
+        s_log(LOG_DEBUG, "No limit detected for the number of clients");
     }
 }
 
@@ -370,7 +379,7 @@ static void daemonize(void) { /* go to background */
 
 static void create_pid(void) {
     int pf;
-    char pid[STRLEN];
+    char *pid;
 
     if(!global_options.pidfile) {
         s_log(LOG_DEBUG, "No pid file being created");
@@ -385,22 +394,24 @@ static void create_pid(void) {
 
     /* silently remove old pid file */
     unlink(global_options.pidfile);
-    if((pf=open(global_options.pidfile, O_WRONLY|O_CREAT|O_TRUNC|O_EXCL,0644))==-1) {
+    pf=open(global_options.pidfile, O_WRONLY|O_CREAT|O_TRUNC|O_EXCL, 0644);
+    if(pf==-1) {
         s_log(LOG_ERR, "Cannot create pid file %s", global_options.pidfile);
         ioerror("create");
         die(1);
     }
-    sprintf(pid, "%lu\n", global_options.dpid);
+    pid=str_printf("%lu\n", global_options.dpid);
     write(pf, pid, strlen(pid));
+    str_free(pid);
     close(pf);
     s_log(LOG_DEBUG, "Created pid file %s", global_options.pidfile);
     atexit(delete_pid);
 }
 
 static void delete_pid(void) {
-    s_log(LOG_DEBUG, "removing pid file %s", global_options.pidfile);
     if((unsigned long)getpid()!=global_options.dpid)
         return; /* current process is not main daemon process */
+    s_log(LOG_DEBUG, "removing pid file %s", global_options.pidfile);
     if(unlink(global_options.pidfile)<0)
         ioerror(global_options.pidfile); /* not critical */
 }
@@ -497,11 +508,9 @@ int s_pipe(int pipefd[2], int nonblock, char *msg) {
 #endif
 
 static int setup_fd(int fd, int nonblock, char *msg) {
-#ifdef USE_WIN32
-    unsigned long l;
-#else /* USE_WIN32 */
-    int err, flags;
-#endif /* USE_WIN32 */
+#ifdef FD_CLOEXEC
+    int err;
+#endif /* FD_CLOEXEC */
 
     if(fd<0) {
         sockerror(msg);
@@ -514,96 +523,115 @@ static int setup_fd(int fd, int nonblock, char *msg) {
         return -1;
     }
 #ifndef USE_NEW_LINUX_API
-#if defined F_GETFL && defined F_SETFL && defined O_NONBLOCK && !defined __INNOTEK_LIBC__
-    do {
-        flags=fcntl(fd, F_GETFL, 0);
-    } while(flags<0 && get_last_socket_error()==EINTR);
-    if(nonblock)
-        flags|=O_NONBLOCK;
-    else
-        flags&=~O_NONBLOCK;
+    set_nonblock(fd, nonblock);
 #ifdef FD_CLOEXEC
-    flags|=FD_CLOEXEC;
-#endif /* FD_CLOEXEC */
     do {
-        err=fcntl(fd, F_SETFL, flags);
+        err=fcntl(fd, F_SETFD, FD_CLOEXEC);
     } while(err<0 && get_last_socket_error()==EINTR);
     if(err<0)
-        sockerror("fcntl"); /* non-critical */
-#else /* use fcntl() */
-    if(ioctlsocket(fd, FIONBIO, &l)<0)
-        sockerror("ioctlsocket"); /* non-critical */
-#endif /* use fcntl() */
+        sockerror("fcntl SETFD"); /* non-critical */
+#endif /* FD_CLOEXEC */
 #endif /* USE_NEW_LINUX_API */
     s_log(LOG_DEBUG, "%s: FD=%d allocated (%sblocking mode)",
         msg, fd, nonblock?"non-":"");
     return fd;
 }
 
+void set_nonblock(int fd, unsigned long nonblock) {
+#if defined F_GETFL && defined F_SETFL && defined O_NONBLOCK && !defined __INNOTEK_LIBC__
+    int err, flags;
+
+    do {
+        flags=fcntl(fd, F_GETFL, 0);
+    } while(flags<0 && get_last_socket_error()==EINTR);
+    if(flags<0) {
+        sockerror("fcntl GETFL"); /* non-critical */
+        return;
+    }
+    if(nonblock)
+        flags|=O_NONBLOCK;
+    else
+        flags&=~O_NONBLOCK;
+    do {
+        err=fcntl(fd, F_SETFL, flags);
+    } while(err<0 && get_last_socket_error()==EINTR);
+    if(err<0)
+        sockerror("fcntl SETFL"); /* non-critical */
+#else /* use fcntl() */
+    if(ioctlsocket(fd, FIONBIO, &nonblock)<0)
+        sockerror("ioctlsocket"); /* non-critical */
+#endif /* use fcntl() */
+}
+
 /**************************************** log messages to identify  build */
 
 void stunnel_info(int level) {
-    char line[STRLEN];
-
-    s_log(level, "stunnel " STUNNEL_VERSION " on " HOST " with %s",
-        SSLeay_version(SSLEAY_VERSION));
-
-    safecopy(line, "Threading:");
+    s_log(level, "stunnel " STUNNEL_VERSION " on " HOST " platform");
+    if(SSLeay()==SSLEAY_VERSION_NUMBER) {
+        s_log(level, "Compiled/running with " OPENSSL_VERSION_TEXT);
+    } else {
+        s_log(level, "Compiled with " OPENSSL_VERSION_TEXT);
+        s_log(level, "Running  with %s", SSLeay_version(SSLEAY_VERSION));
+        s_log(level, "Update OpenSSL shared libraries or rebuild stunnel");
+    }
+    s_log(level,
+        "Threading:"
 #ifdef USE_UCONTEXT
-    safeconcat(line, "UCONTEXT");
+        "UCONTEXT"
 #endif
 #ifdef USE_PTHREAD
-    safeconcat(line, "PTHREAD");
+        "PTHREAD"
 #endif
 #ifdef USE_WIN32
-    safeconcat(line, "WIN32");
+        "WIN32"
 #endif
 #ifdef USE_FORK
-    safeconcat(line, "FORK");
+        "FORK"
 #endif
 
-    safeconcat(line, " SSL:");
+        " SSL:"
 #ifdef HAVE_OSSL_ENGINE_H
-    safeconcat(line, "ENGINE");
+        "ENGINE"
 #else /* defined(HAVE_OSSL_ENGINE_H) */
-    safeconcat(line, "NOENGINE");
+        "NOENGINE"
 #endif /* defined(HAVE_OSSL_ENGINE_H) */
 #ifdef USE_FIPS
-    safeconcat(line, ",FIPS");
+        ",FIPS"
 #endif /* USE_FIPS */
 
-    safeconcat(line, " Sockets:");
-#ifdef USE_POLL
-    safeconcat(line, "POLL");
-#else /* defined(USE_POLL) */
-    safeconcat(line, "SELECT");
-#endif /* defined(USE_POLL) */
-#if defined(USE_WIN32) && !defined(_WIN32_WCE)
-    if(s_getaddrinfo)
-        safeconcat(line, ",IPv6");
-    else
-        safeconcat(line, ",IPv4");
-#else /* defined(USE_WIN32) */
-#if defined(USE_IPv6)
-    safeconcat(line, ",IPv6");
-#else /* defined(USE_IPv6) */
-    safeconcat(line, ",IPv4");
-#endif /* defined(USE_IPv6) */
-#endif /* defined(USE_WIN32) */
-
+        " Auth:"
 #ifdef USE_LIBWRAP
-    safeconcat(line, " Auth:LIBWRAP");
+        "LIBWRAP"
+#else
+        "none"
 #endif
 
-    s_log(level, "%s", line);
+        " Sockets:"
+#ifdef USE_POLL
+        "POLL"
+#else /* defined(USE_POLL) */
+        "SELECT"
+#endif /* defined(USE_POLL) */
+        ",IPv%c",
+#if defined(USE_WIN32) && !defined(_WIN32_WCE)
+        s_getaddrinfo ? '6' : '4'
+#else /* defined(USE_WIN32) */
+#if defined(USE_IPv6)
+        '6'
+#else /* defined(USE_IPv6) */
+        '4'
+#endif /* defined(USE_IPv6) */
+#endif /* defined(USE_WIN32) */
+        );
 }
 
 /**************************************** fatal error */
 
 void die(int status) { /* some cleanup and exit */
+    str_stats();
     log_flush(LOG_MODE_ERROR);
 #ifdef USE_WIN32
-    exit_win32(status);
+    win_exit(status);
 #else
     exit(status);
 #endif
