@@ -1,6 +1,6 @@
 /*
  *   stunnel       Universal SSL tunnel
- *   Copyright (C) 1998-2009 Michal Trojnara <Michal.Trojnara@mirt.net>
+ *   Copyright (C) 1998-2011 Michal Trojnara <Michal.Trojnara@mirt.net>
  *
  *   This program is free software; you can redistribute it and/or modify it
  *   under the terms of the GNU General Public License as published by the
@@ -41,34 +41,34 @@
 /**************************************** prototypes */
 
 /* verify initialization */
-static void load_file_lookup(X509_STORE *, char *);
-static void add_dir_lookup(X509_STORE *, char *);
+static int load_file_lookup(X509_STORE *, char *);
+static int add_dir_lookup(X509_STORE *, char *);
 
 /* verify callback */
 static int verify_callback(int, X509_STORE_CTX *);
-static int cert_check(CLI *c, X509_STORE_CTX *, char *, int);
-static int crl_check(CLI *c, X509_STORE_CTX *, char *);
-static int ocsp_check(CLI *c, X509_STORE_CTX *, char *);
+static int cert_check(CLI *c, X509_STORE_CTX *, int);
+static int crl_check(CLI *c, X509_STORE_CTX *);
+static int ocsp_check(CLI *c, X509_STORE_CTX *);
 
 /* utility functions */
 static void log_time(const int, const char *, ASN1_TIME *);
 
 /**************************************** verify initialization */
 
-void verify_init(LOCAL_OPTIONS *section) {
+int verify_init(SERVICE_OPTIONS *section) {
     if(section->verify_level<0)
-        return; /* No certificate verification */
+        return 0; /* OK - no certificate verification */
 
-    if(section->verify_level>1 && !section->ca_file && !section->ca_dir) {
+    if(section->verify_level>=2 && !section->ca_file && !section->ca_dir) {
         s_log(LOG_ERR,
             "Either CApath or CAfile has to be used for authentication");
-        die(1);
+        return 1; /* FAILED */
     }
 
     section->revocation_store=X509_STORE_new();
     if(!section->revocation_store) {
         sslerror("X509_STORE_new");
-        die(1);
+        return 1; /* FAILED */
     }
 
     if(section->ca_file) {
@@ -77,14 +77,15 @@ void verify_init(LOCAL_OPTIONS *section) {
             s_log(LOG_ERR, "Error loading verify certificates from %s",
                 section->ca_file);
             sslerror("SSL_CTX_load_verify_locations");
-            die(1);
+            return 1; /* FAILED */
         }
         /* list of trusted CAs for the client to choose the right cert */
         SSL_CTX_set_client_CA_list(section->ctx,
             SSL_load_client_CA_file(section->ca_file));
         s_log(LOG_DEBUG, "Loaded verify certificates from %s",
             section->ca_file);
-        load_file_lookup(section->revocation_store, section->ca_file);
+        if(load_file_lookup(section->revocation_store, section->ca_file))
+            return 1; /* FAILED */
     }
 
     if(section->ca_dir) {
@@ -93,57 +94,62 @@ void verify_init(LOCAL_OPTIONS *section) {
             s_log(LOG_ERR, "Error setting verify directory to %s",
                 section->ca_dir);
             sslerror("SSL_CTX_load_verify_locations");
-            die(1);
+            return 1; /* FAILED */
         }
         s_log(LOG_DEBUG, "Verify directory set to %s", section->ca_dir);
         add_dir_lookup(section->revocation_store, section->ca_dir);
     }
 
     if(section->crl_file)
-        load_file_lookup(section->revocation_store, section->crl_file);
+        if(load_file_lookup(section->revocation_store, section->crl_file))
+            return 1; /* FAILED */
 
     if(section->crl_dir) {
         section->revocation_store->cache=0; /* don't cache CRLs */
         add_dir_lookup(section->revocation_store, section->crl_dir);
     }
 
-    SSL_CTX_set_verify(section->ctx, section->verify_level==SSL_VERIFY_NONE ?
-        SSL_VERIFY_PEER : section->verify_level, verify_callback);
+    SSL_CTX_set_verify(section->ctx, SSL_VERIFY_PEER |
+        (section->verify_level>=2 ? SSL_VERIFY_FAIL_IF_NO_PEER_CERT : 0),
+        verify_callback);
 
-    if(section->ca_dir && section->verify_use_only_my)
-        s_log(LOG_NOTICE, "Peer certificate location %s", section->ca_dir);
+    if(section->ca_dir && section->verify_level>=3)
+        s_log(LOG_INFO, "Peer certificate location %s", section->ca_dir);
+    return 0; /* OK */
 }
 
-static void load_file_lookup(X509_STORE *store, char *name) {
+static int load_file_lookup(X509_STORE *store, char *name) {
     X509_LOOKUP *lookup;
 
     lookup=X509_STORE_add_lookup(store, X509_LOOKUP_file());
     if(!lookup) {
         sslerror("X509_STORE_add_lookup");
-        die(1);
+        return 1; /* FAILED */
     }
     if(!X509_LOOKUP_load_file(lookup, name, X509_FILETYPE_PEM)) {
         s_log(LOG_ERR, "Failed to load %s revocation lookup file", name);
         sslerror("X509_LOOKUP_load_file");
-        die(1);
+        return 1; /* FAILED */
     }
     s_log(LOG_DEBUG, "Loaded %s revocation lookup file", name);
+    return 0; /* OK */
 }
 
-static void add_dir_lookup(X509_STORE *store, char *name) {
+static int add_dir_lookup(X509_STORE *store, char *name) {
     X509_LOOKUP *lookup;
 
     lookup=X509_STORE_add_lookup(store, X509_LOOKUP_hash_dir());
     if(!lookup) {
         sslerror("X509_STORE_add_lookup");
-        die(1);
+        return 1; /* FAILED */
     }
     if(!X509_LOOKUP_add_dir(lookup, name, X509_FILETYPE_PEM)) {
         s_log(LOG_ERR, "Failed to add %s revocation lookup directory", name);
         sslerror("X509_LOOKUP_add_dir");
-        die(1);
+        return 1; /* FAILED */
     }
     s_log(LOG_DEBUG, "Added %s revocation lookup directory", name);
+    return 0; /* OK */
 }
 
 /**************************************** verify callback */
@@ -152,58 +158,74 @@ static int verify_callback(int preverify_ok, X509_STORE_CTX *callback_ctx) {
         /* our verify callback function */
     SSL *ssl;
     CLI *c;
-    char subject_name[STRLEN];
+    char *subject_name;
 
-    X509_NAME_oneline(X509_get_subject_name(callback_ctx->current_cert),
-        subject_name, STRLEN);
-    safestring(subject_name);
-
-    /* retrieve the pointer to the SSL of the connection currently treated
-     * and the application specific data stored into the SSL object */
+    /* retrieve application specific data */
     ssl=X509_STORE_CTX_get_ex_data(callback_ctx,
         SSL_get_ex_data_X509_STORE_CTX_idx());
     c=SSL_get_ex_data(ssl, cli_index);
 
-    if(!cert_check(c, callback_ctx, subject_name, preverify_ok))
-        return 0; /* reject connection */
-    if(!crl_check(c, callback_ctx, subject_name))
-        return 0; /* reject connection */
-#if SSLEAY_VERSION_NUMBER >= 0x00907000L
-    if(c->opt->option.ocsp && !ocsp_check(c, callback_ctx, subject_name))
-        return 0; /* reject connection */
-#endif /* OpenSSL-0.9.7 */
+    /* certificate name for logging */
+    subject_name=X509_NAME_oneline(
+        X509_get_subject_name(callback_ctx->current_cert), NULL, 0);
 
-    /* errnum=X509_STORE_CTX_get_error(ctx); */
-    s_log(LOG_NOTICE, "VERIFY OK: depth=%d, %s",
+    s_log(LOG_DEBUG, "Starting certificate verification: depth=%d, %s",
         callback_ctx->error_depth, subject_name);
+    if(!cert_check(c, callback_ctx, preverify_ok)) {
+        s_log(LOG_WARNING, "Certificate check failed: depth=%d, %s",
+            callback_ctx->error_depth, subject_name);
+        OPENSSL_free(subject_name);
+        return 0; /* reject connection */
+    }
+    if(!crl_check(c, callback_ctx)) {
+        s_log(LOG_WARNING, "CRL check failed: depth=%d, %s",
+            callback_ctx->error_depth, subject_name);
+        OPENSSL_free(subject_name);
+        return 0; /* reject connection */
+    }
+    if(c->opt->option.ocsp && !ocsp_check(c, callback_ctx)) {
+        s_log(LOG_WARNING, "OCSP check failed: depth=%d, %s",
+            callback_ctx->error_depth, subject_name);
+        OPENSSL_free(subject_name);
+        return 0; /* reject connection */
+    }
+    /* errnum=X509_STORE_CTX_get_error(ctx); */
+    s_log(LOG_NOTICE, "Certificate accepted: depth=%d, %s",
+        callback_ctx->error_depth, subject_name);
+    OPENSSL_free(subject_name);
     return 1; /* accept connection */
 }
 
 /**************************************** certificate checking */
 
-static int cert_check(CLI *c, X509_STORE_CTX *callback_ctx,
-        char *subject_name, int preverify_ok) {
-    X509_OBJECT ret;
+static int cert_check(CLI *c, X509_STORE_CTX *callback_ctx, int preverify_ok) {
+    X509_OBJECT obj;
+    ASN1_BIT_STRING *local_key, *peer_key;
 
-    if(c->opt->verify_level==SSL_VERIFY_NONE) {
-        s_log(LOG_INFO, "VERIFY IGNORE: depth=%d, %s",
-            callback_ctx->error_depth, subject_name);
+    if(c->opt->verify_level<1) {
+        s_log(LOG_INFO, "CERT: Verification not enabled");
         return 1; /* accept connection */
     }
     if(!preverify_ok) {
         /* remote site specified a certificate, but it's not correct */
-        s_log(LOG_WARNING, "VERIFY ERROR: depth=%d, error=%s: %s",
-            callback_ctx->error_depth,
-            X509_verify_cert_error_string (callback_ctx->error),
-                subject_name);
+        s_log(LOG_WARNING, "CERT: Verification error: %s",
+            X509_verify_cert_error_string(callback_ctx->error));
         return 0; /* reject connection */
     }
-    if(c->opt->verify_use_only_my && callback_ctx->error_depth==0 &&
-            X509_STORE_get_by_subject(callback_ctx, X509_LU_X509,
-                X509_get_subject_name(callback_ctx->current_cert), &ret)!=1) {
-        s_log(LOG_WARNING, "VERIFY ERROR ONLY MY: no cert for %s",
-            subject_name);
-        return 0; /* reject connection */
+    if(c->opt->verify_level>=3 && callback_ctx->error_depth==0) {
+        if(X509_STORE_get_by_subject(callback_ctx, X509_LU_X509,
+                X509_get_subject_name(callback_ctx->current_cert), &obj)!=1) {
+            s_log(LOG_WARNING, "CERT: Certificate not found in local repository");
+            return 0; /* reject connection */
+        }
+        peer_key=X509_get0_pubkey_bitstr(callback_ctx->current_cert);
+        local_key=X509_get0_pubkey_bitstr(obj.data.x509);
+        if(!peer_key || !local_key || peer_key->length!=local_key->length ||
+                memcmp(peer_key->data, local_key->data, local_key->length)) {
+            s_log(LOG_WARNING, "CERT: Public keys do not match");
+            return 0; /* reject connection */
+        }
+        s_log(LOG_INFO, "CERT: Locally installed certificate matched");
     }
     return 1; /* accept connection */
 }
@@ -211,8 +233,7 @@ static int cert_check(CLI *c, X509_STORE_CTX *callback_ctx,
 /**************************************** CRL checking */
 
 /* based on BSD-style licensed code of mod_ssl */
-static int crl_check(CLI *c, X509_STORE_CTX *callback_ctx,
-        char *subject_name) {
+static int crl_check(CLI *c, X509_STORE_CTX *callback_ctx) {
     X509_STORE_CTX store_ctx;
     X509_OBJECT obj;
     X509_NAME *subject;
@@ -250,7 +271,7 @@ static int crl_check(CLI *c, X509_STORE_CTX *callback_ctx,
         /* verify the signature on this CRL */
         pubkey=X509_get_pubkey(cert);
         if(X509_CRL_verify(crl, pubkey)<=0) {
-            s_log(LOG_WARNING, "Invalid signature on CRL");
+            s_log(LOG_WARNING, "CRL: Invalid signature");
             X509_STORE_CTX_set_error(callback_ctx,
                 X509_V_ERR_CRL_SIGNATURE_FAILURE);
             X509_OBJECT_free_contents(&obj);
@@ -263,15 +284,14 @@ static int crl_check(CLI *c, X509_STORE_CTX *callback_ctx,
 
         /* check date of CRL to make sure it's not expired */
         if(!next_update) {
-            s_log(LOG_WARNING, "Found CRL has invalid nextUpdate field");
+            s_log(LOG_WARNING, "CRL: Invalid nextUpdate field");
             X509_STORE_CTX_set_error(callback_ctx,
                 X509_V_ERR_ERROR_IN_CRL_NEXT_UPDATE_FIELD);
             X509_OBJECT_free_contents(&obj);
             return 0; /* reject connection */
         }
         if(X509_cmp_current_time(next_update)<0) {
-            s_log(LOG_WARNING, "Found CRL is expired - "
-                "revoking all certificates until you get updated CRL");
+            s_log(LOG_WARNING, "CRL: CRL Expired - revoking all certificates");
             X509_STORE_CTX_set_error(callback_ctx, X509_V_ERR_CRL_HAS_EXPIRED);
             X509_OBJECT_free_contents(&obj);
             return 0; /* reject connection */
@@ -288,22 +308,14 @@ static int crl_check(CLI *c, X509_STORE_CTX *callback_ctx,
     crl=obj.data.crl;
     if(rc>0 && crl) {
         /* check if the current certificate is revoked by this CRL */
-#if SSLEAY_VERSION_NUMBER >= 0x00904000
         n=sk_X509_REVOKED_num(X509_CRL_get_REVOKED(crl));
-#else
-        n=sk_num(X509_CRL_get_REVOKED(crl));
-#endif
         for(i=0; i<n; i++) {
-#if SSLEAY_VERSION_NUMBER >= 0x00904000
             revoked=sk_X509_REVOKED_value(X509_CRL_get_REVOKED(crl), i);
-#else
-            revoked=(X509_REVOKED *)sk_value(X509_CRL_get_REVOKED(crl), i);
-#endif
             if(ASN1_INTEGER_cmp(revoked->serialNumber,
                     X509_get_serialNumber(cert)) == 0) {
                 serial=ASN1_INTEGER_get(revoked->serialNumber);
                 cp=X509_NAME_oneline(issuer, NULL, 0);
-                s_log(LOG_WARNING, "Certificate with serial %ld (0x%lX) "
+                s_log(LOG_WARNING, "CRL: Certificate with serial %ld (0x%lX) "
                     "revoked per CRL from issuer %s", serial, serial, cp);
                 OPENSSL_free(cp);
                 X509_STORE_CTX_set_error(callback_ctx, X509_V_ERR_CERT_REVOKED);
@@ -313,15 +325,13 @@ static int crl_check(CLI *c, X509_STORE_CTX *callback_ctx,
         }
         X509_OBJECT_free_contents(&obj);
     }
-    s_log(LOG_NOTICE, "CRL: verification passed");
     return 1; /* accept connection */
 }
 
 /**************************************** OCSP checking */
+/* TODO: check OCSP server specified in the certificate */
 
-#if SSLEAY_VERSION_NUMBER >= 0x00907000L
-static int ocsp_check(CLI *c, X509_STORE_CTX *callback_ctx,
-        char *subject_name) {
+static int ocsp_check(CLI *c, X509_STORE_CTX *callback_ctx) {
     int error, retval=0;
     SOCKADDR_UNION addr;
     X509 *cert;
@@ -335,17 +345,11 @@ static int ocsp_check(CLI *c, X509_STORE_CTX *callback_ctx,
         *this_update=NULL, *next_update=NULL;
     int status, reason;
 
-    /* TODO: check OCSP server specified in the certificate */
-    s_log(LOG_DEBUG, "OCSP: starting verification");
-
     /* connect specified OCSP server (responder) */
-    if((c->fd=
-        socket(c->opt->ocsp_addr.addr[0].sa.sa_family, SOCK_STREAM, 0))<0) {
-        sockerror("socket (auth_user)");
+    c->fd=s_socket(c->opt->ocsp_addr.addr[0].sa.sa_family, SOCK_STREAM, 0,
+        0, "OCSP: socket (auth_user)");
+    if(c->fd<0)
         return 0; /* reject connection */
-    }
-    if(alloc_fd(c->fd))
-        goto cleanup;
     memcpy(&addr, &c->opt->ocsp_addr.addr[0], sizeof addr);
     if(connect_blocking(c, &addr, addr_len(addr)))
         goto cleanup;
@@ -354,23 +358,23 @@ static int ocsp_check(CLI *c, X509_STORE_CTX *callback_ctx,
     /* get current certificate ID */
     cert=X509_STORE_CTX_get_current_cert(callback_ctx); /* get current cert */
     if(X509_STORE_CTX_get1_issuer(&issuer, callback_ctx, cert)!=1) {
-        sslerror("X509_STORE_CTX_get1_issuer");
+        sslerror("OCSP: X509_STORE_CTX_get1_issuer");
         goto cleanup;
     }
     certID=OCSP_cert_to_id(0, cert, issuer);
     if(!certID) {
-        sslerror("OCSP_cert_to_id");
+        sslerror("OCSP: OCSP_cert_to_id");
         goto cleanup;
     }
 
     /* build request */
     request=OCSP_REQUEST_new();
     if(!request) {
-        sslerror("OCSP_REQUEST_new");
+        sslerror("OCSP: OCSP_REQUEST_new");
         goto cleanup;
     }
     if(!OCSP_request_add0_id(request, certID)) {
-        sslerror("OCSP_request_add0_id");
+        sslerror("OCSP: OCSP_request_add0_id");
         goto cleanup;
     }
     OCSP_request_add1_nonce(request, 0, -1);
@@ -379,62 +383,57 @@ static int ocsp_check(CLI *c, X509_STORE_CTX *callback_ctx,
     /* FIXME: this code won't work with ucontext threading */
     /* (blocking sockets are used) */
     bio=BIO_new_fd(c->fd, BIO_NOCLOSE);
-    setnonblock(c->fd, 0);
     response=OCSP_sendreq_bio(bio, c->opt->ocsp_path, request);
-    setnonblock(c->fd, 1);
     if(!response) {
-        sslerror("OCSP_sendreq_bio");
+        sslerror("OCSP: OCSP_sendreq_bio");
         goto cleanup;
     }
     error=OCSP_response_status(response);
     if(error!=OCSP_RESPONSE_STATUS_SUCCESSFUL) {
-        s_log(LOG_WARNING, "OCSP: responder error: %d: %s",
+        s_log(LOG_WARNING, "OCSP: Responder error: %d: %s",
             error, OCSP_response_status_str(error));
         goto cleanup;
     }
-    s_log(LOG_DEBUG, "OCSP: response received");
+    s_log(LOG_DEBUG, "OCSP: Response received");
 
     /* verify the response */
     basicResponse=OCSP_response_get1_basic(response);
     if(!basicResponse) {
-        sslerror("OCSP_response_get1_basic");
+        sslerror("OCSP: OCSP_response_get1_basic");
         goto cleanup;
     }
     if(OCSP_check_nonce(request, basicResponse)<=0) {
-        sslerror("OCSP_check_nonce");
+        sslerror("OCSP: OCSP_check_nonce");
         goto cleanup;
     }
     if(OCSP_basic_verify(basicResponse, NULL,
             c->opt->revocation_store, c->opt->ocsp_flags)<=0) {
-        sslerror("OCSP_basic_verify");
+        sslerror("OCSP: OCSP_basic_verify");
         goto cleanup;
     }
     if(!OCSP_resp_find_status(basicResponse, certID, &status, &reason,
             &revoked_at, &this_update, &next_update)) {
-        sslerror("OCSP_resp_find_status");
+        sslerror("OCSP: OCSP_resp_find_status");
         goto cleanup;
     }
-    s_log(LOG_NOTICE, "OCSP: status: %d: %s",
+    s_log(LOG_NOTICE, "OCSP: Status: %d: %s",
         status, OCSP_cert_status_str(status));
-    log_time(LOG_INFO, "OCSP: this update", this_update);
-    log_time(LOG_INFO, "OCSP: next update", next_update);
+    log_time(LOG_INFO, "OCSP: This update", this_update);
+    log_time(LOG_INFO, "OCSP: Next update", next_update);
     /* check if the response is valid for at least one minute */
     if(!OCSP_check_validity(this_update, next_update, 60, -1)) {
-        sslerror("OCSP_check_validity");
+        sslerror("OCSP: OCSP_check_validity");
         goto cleanup;
     }
     if(status==V_OCSP_CERTSTATUS_REVOKED) {
         if(reason==-1)
-            s_log(LOG_WARNING, "OCSP: certificate revoked");
+            s_log(LOG_WARNING, "OCSP: Certificate revoked");
         else
-            s_log(LOG_WARNING, "OCSP: certificate revoked: %d: %s",
+            s_log(LOG_WARNING, "OCSP: Certificate revoked: %d: %s",
                 reason, OCSP_crl_reason_str(reason));
-        log_time(LOG_NOTICE, "OCSP: revoked at", revoked_at);
+        log_time(LOG_NOTICE, "OCSP: Revoked at", revoked_at);
         goto cleanup;
     }
-
-    /* success */
-    s_log(LOG_NOTICE, "OCSP: verification passed");
     retval=1; /* accept connection */
 cleanup:
     if(bio)
@@ -451,7 +450,6 @@ cleanup:
     c->fd=-1; /* avoid double close on cleanup */
     return retval;
 }
-#endif /* OpenSSL-0.9.7 */
 
 static void log_time(const int level, const char *txt, ASN1_TIME *t) {
     char *cp;
@@ -461,14 +459,25 @@ static void log_time(const int level, const char *txt, ASN1_TIME *t) {
     if(!t)
         return;
     bio=BIO_new(BIO_s_mem());
+    if(!bio)
+        return;
     ASN1_TIME_print(bio, t);
     n=BIO_pending(bio);
-    cp=malloc(n+1);
+    cp=str_alloc(n+1);
+    if(!cp) {
+        BIO_free(bio);
+        return;
+    }
     n=BIO_read(bio, cp, n);
+    if(n<0) {
+        BIO_free(bio);
+        str_free(cp);
+        return;
+    }
     cp[n]='\0';
     BIO_free(bio);
     s_log(level, "%s: %s", txt, cp);
-    free(cp);
+    str_free(cp);
 }
 
-/* End of verify.c */
+/* end of verify.c */
